#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Modis level 3 hdf-eos format reader.

Introduction
------------

The ``modis_l3`` reader reads MODIS L3 products in HDF-EOS format.

There are multiple level 3 products, including some on sinusoidal grids and some on the climate modeling grid (CMG).
This reader supports the CMG products at present, and the sinusoidal products will be added if there is demand.

The reader has been tested with:
    - MCD43c*: BRDF/Albedo data, such as parameters, albedo and nbar
    - MOD09CMG: Surface Reflectance on climate monitoring grid.

To get a list of the available datasets for a given file refer to the "Load data" section in :doc:`../reading`.

"""
import logging
from typing import Iterable

from pyresample import geometry

from satpy.readers.core.hdfeos import HDFEOSGeoReader

logger = logging.getLogger(__name__)


class ModisL3GriddedHDFFileHandler(HDFEOSGeoReader):
    """File handler for MODIS HDF-EOS Level 3 CMG gridded files."""
    def available_datasets(self, configured_datasets=None):
        """Automatically determine datasets provided by this file."""
        # Initialise set of variable names to carry through code
        handled_var_names = set()

        ds_dict = self.sd.datasets()

        for is_avail, ds_info in (configured_datasets or []):
            file_key = ds_info.get("file_key", ds_info["name"])
            # we must add all variables here even if another file handler has
            # claimed the variable. It could be another instance of this file
            # type, and we don't want to add that variable dynamically if the
            # other file handler defined it by the YAML definition.
            handled_var_names.add(file_key)
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info
                continue
            if self.file_type_matches(ds_info["file_type"]) is None:
                # this is not the file type for this dataset
                yield None, ds_info
                continue
            yield file_key in ds_dict.keys(), ds_info

        yield from self._dynamic_variables_from_file(handled_var_names)

    def _dynamic_variables_from_file(self, handled_var_names: set) -> Iterable[tuple[bool, dict]]:
        res = self._get_res()
        for var_name in self.sd.datasets().keys():
            if var_name in handled_var_names:
                # skip variables that YAML had configured
                continue
            common = {"file_type": "modis_l3_cmg_hdf",
                      "resolution": res,
                      "name": var_name}
            yield True, common


    def _get_res(self):
        """Compute the resolution from the file metadata."""
        gridname = self.metadata["GridStructure"]["GRID_1"]["GridName"]
        if "CMG" not in gridname:
            raise ValueError("Only CMG grids are supported")

        # Get the grid resolution from the grid name
        pos = gridname.rfind("_") + 1
        pos2 = gridname.rfind("Deg")

        # Initialise number of rows and columns
        # Some products don't have resolution listed.
        if pos < 0 or pos2 < 0:
            return 360. / self.metadata["GridStructure"]["GRID_1"]["XDim"]
        else:
            return float(gridname[pos:pos2])

    def get_dataset(self, dataset_id, dataset_info):
        """Get DataArray for specified dataset."""
        dataset_name = dataset_id["name"]
        dataset = self.load_dataset(dataset_name, dataset_info.pop("category", False))
        self._add_satpy_metadata(dataset_id, dataset)

        return dataset

    def _get_area_extent(self):
        """Get the grid properties."""
        # Now compute the data extent
        upperleft = self.metadata["GridStructure"]["GRID_1"]["UpperLeftPointMtrs"]
        lowerright = self.metadata["GridStructure"]["GRID_1"]["LowerRightMtrs"]

        # For some reason, a few of the CMG products multiply their
        # decimal degree extents by one million. This fixes it.
        if lowerright[0] > 1e6 or upperleft[0] > 1e6:
            upperleft = tuple(val / 1e6 for val in upperleft)
            lowerright = tuple(val / 1e6 for val in lowerright)

        return upperleft[0], lowerright[1], lowerright[0], upperleft[1]

    def get_area_def(self, dsid):
        """Get the area definition.

        This is fixed, but not defined in the file. So we must
        generate it ourselves with some assumptions.
        """
        proj_param = "EPSG:4326"

        # Get the size of the dataset
        nrows = self.metadata["GridStructure"]["GRID_1"]["YDim"]
        ncols = self.metadata["GridStructure"]["GRID_1"]["XDim"]

        # Construct the area definition
        area = geometry.AreaDefinition("gridded_modis",
                                       "A gridded L3 MODIS area",
                                       "longlat",
                                       proj_param,
                                       ncols,
                                       nrows,
                                       self._get_area_extent())

        return area



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface for BaseFileHandlers."""

import numpy as np
import xarray as xr
from pyresample.geometry import SwathDefinition

from satpy.dataset import combine_metadata
from satpy.readers.core.remote import open_file_or_filename


def open_dataset(filename, *args, **kwargs):  # noqa: D417
    """Open a file with xarray.

    Args:
       filename (Union[str, FSFile]):
           The path to the file to open. Can be a `string` or
           :class:`~satpy.readers.core.remote.FSFile` object which allows using
           `fsspec` or `s3fs` like files.

    Returns:
       xarray.Dataset:

    Notes:
       This can be used to enable readers to open remote files.
    """
    f_obj = open_file_or_filename(filename)
    return xr.open_dataset(f_obj, *args, **kwargs)


class BaseFileHandler:
    """Base file handler."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize file handler."""
        self.filename = filename
        self.navigation_reader = None
        self.filename_info = filename_info
        self.filetype_info = filetype_info
        self.metadata = filename_info.copy()

    def __str__(self):
        """Customize __str__."""
        return "<{}: '{}'>".format(self.__class__.__name__, self.filename)

    def __repr__(self):
        """Customize __repr__."""
        return str(self)

    def get_dataset(self, dataset_id, ds_info):
        """Get dataset."""
        raise NotImplementedError

    def get_area_def(self, dsid):
        """Get area definition."""
        raise NotImplementedError

    def get_bounding_box(self):
        """Get the bounding box of the files, as a (lons, lats) tuple.

        The tuple return should a lons and lats list of coordinates traveling
        clockwise around the points available in the file.
        """
        raise NotImplementedError

    @staticmethod
    def _combine(infos, func, *keys):
        res = {}
        for key in keys:
            if key in infos[0]:
                res[key] = func([i[key] for i in infos])

        return res

    def combine_info(self, all_infos):
        """Combine metadata for multiple datasets.

        When loading data from multiple files it can be non-trivial to combine
        things like start_time, end_time, start_orbit, end_orbit, etc.

        By default this method will produce a dictionary containing all values
        that were equal across **all** provided info dictionaries.

        Additionally it performs the logical comparisons to produce the
        following if they exist:

         - start_time
         - end_time
         - start_orbit
         - end_orbit
         - orbital_parameters
         - time_parameters

         Also, concatenate the areas.

        """
        combined_info = combine_metadata(*all_infos)

        new_dict = self._combine(all_infos, min, "start_orbit")
        new_dict.update(self._combine(all_infos, max, "end_orbit"))
        new_dict.update(self._combine_orbital_parameters(all_infos))

        try:
            area = SwathDefinition(lons=np.ma.vstack([info["area"].lons for info in all_infos]),
                                   lats=np.ma.vstack([info["area"].lats for info in all_infos]))
            area.name = "_".join([info["area"].name for info in all_infos])
            combined_info["area"] = area
        except KeyError:
            pass

        new_dict.update(combined_info)
        return new_dict

    def _combine_orbital_parameters(self, all_infos):
        orb_params = [info.get("orbital_parameters", {}) for info in all_infos]
        if not all(orb_params):
            return {}
        # Collect all available keys
        orb_params_comb = {}
        for d in orb_params:
            orb_params_comb.update(d)

        # Average known keys
        keys = ["projection_longitude", "projection_latitude", "projection_altitude",
                "satellite_nominal_longitude", "satellite_nominal_latitude",
                "satellite_actual_longitude", "satellite_actual_latitude", "satellite_actual_altitude",
                "nadir_longitude", "nadir_latitude"]
        orb_params_comb.update(self._combine(orb_params, np.mean, *keys))
        return {"orbital_parameters": orb_params_comb}

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get end time."""
        return self.filename_info.get("end_time", self.start_time)

    @property
    def sensor_names(self):
        """List of sensors represented in this file."""
        raise NotImplementedError

    def file_type_matches(self, ds_ftype):
        """Match file handler's type to this dataset's file type.

        Args:
            ds_ftype (str or list): File type or list of file types that a
                dataset is configured to be loaded from.

        Returns:
            ``True`` if this file handler object's type matches the
            dataset's file type(s), ``None`` otherwise. ``None`` is returned
            instead of ``False`` to follow the convention of the
            :meth:`available_datasets` method.

        """
        if not isinstance(ds_ftype, (list, tuple)):
            ds_ftype = [ds_ftype]
        if self.filetype_info["file_type"] in ds_ftype:
            return True
        return None

    def available_datasets(self, configured_datasets=None):
        """Get information of available datasets in this file.

        This is used for dynamically specifying what datasets are available
        from a file in addition to what's configured in a YAML configuration
        file. Note that this method is called for each file handler for each
        file type; care should be taken when possible to reduce the amount
        of redundant datasets produced.

        This method should **not** update values of the dataset information
        dictionary **unless** this file handler has a matching file type
        (the data could be loaded from this object in the future) and at least
        **one** :class:`satpy.dataset.dataid.DataID` key is also modified.
        Otherwise, this file type may override the information provided by
        a more preferred file type (as specified in the YAML file).
        It is recommended that any non-ID metadata be updated during the
        :meth:`BaseFileHandler.get_dataset` part of loading.
        This method is not guaranteed that it will be called before any
        other file type's handler.
        The availability "boolean" not being ``None`` does not mean that a
        file handler called later can't provide an additional dataset, but
        it must provide more identifying (DataID) information to do so
        and should yield its new dataset in addition to the previous one.

        Args:
            configured_datasets (list): Series of (bool or None, dict) in the
                same way as is returned by this method (see below). The bool
                is whether the dataset is available from at least one
                of the current file handlers. It can also be ``None`` if
                no file handler before us knows how to handle it.
                The dictionary is existing dataset metadata. The dictionaries
                are typically provided from a YAML configuration file and may
                be modified, updated, or used as a "template" for additional
                available datasets. This argument could be the result of a
                previous file handler's implementation of this method.

        Returns:
            Iterator of (bool or None, dict) pairs where dict is the
            dataset's metadata. If the dataset is available in the current
            file type then the boolean value should be ``True``, ``False``
            if we **know** about the dataset but it is unavailable, or
            ``None`` if this file object is not responsible for it.

        Example 1 - Supplement existing configured information::

            def available_datasets(self, configured_datasets=None):
                "Add information to configured datasets."
                # we know the actual resolution
                res = self.resolution

                # update previously configured datasets
                for is_avail, ds_info in (configured_datasets or []):
                    # some other file handler knows how to load this
                    # don't override what they've done
                    if is_avail is not None:
                        yield is_avail, ds_info

                    matches = self.file_type_matches(ds_info['file_type'])
                    if matches and ds_info.get('resolution') != res:
                        # we are meant to handle this dataset (file type matches)
                        # and the information we can provide isn't available yet
                        new_info = ds_info.copy()
                        new_info['resolution'] = res
                        yield True, new_info
                    elif is_avail is None:
                        # we don't know what to do with this
                        # see if another future file handler does
                        yield is_avail, ds_info

        Example 2 - Add dynamic datasets from the file::

            def available_datasets(self, configured_datasets=None):
                "Add datasets dynamically determined from the file."
                # pass along existing datasets
                for is_avail, ds_info in (configured_datasets or []):
                    if is_avail is not None:
                        # some other file handler said it has this dataset
                        # we don't know any more information than the previous
                        # file handler so let's yield early
                        yield is_avail, ds_info
                        continue
                    yield self.file_type_matches(ds_info["file_type"]), ds_info

                # get dynamic variables known to this file (that we created)
                for var_name, val in self.dynamic_variables.items():
                    ds_info = {
                        'file_type': self.filetype_info['file_type'],
                        'resolution': 1000,
                        'name': var_name,
                    }
                    yield True, ds_info

        """
        for is_avail, ds_info in (configured_datasets or []):
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info
                continue
            yield self.file_type_matches(ds_info["file_type"]), ds_info



"""Bucket resamplers."""

from logging import getLogger

import dask.array as da
import numpy as np
import xarray as xr
from pyresample.resampler import BaseResampler as PRBaseResampler

from satpy.resample.base import _update_resampled_coords
from satpy.utils import get_legacy_chunk_size

LOG = getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()


class BucketResamplerBase(PRBaseResampler):
    """Base class for bucket resampling which implements averaging."""

    def __init__(self, source_geo_def, target_geo_def):
        """Initialize bucket resampler."""
        super(BucketResamplerBase, self).__init__(source_geo_def, target_geo_def)
        self.resampler = None

    def precompute(self, **kwargs):
        """Create X and Y indices and store them for later use."""
        from pyresample import bucket

        LOG.debug("Initializing bucket resampler.")
        source_lons, source_lats = self.source_geo_def.get_lonlats(
            chunks=CHUNK_SIZE)
        self.resampler = bucket.BucketResampler(self.target_geo_def,
                                                source_lons,
                                                source_lats)

    def compute(self, data, **kwargs):
        """Call the resampling."""
        raise NotImplementedError("Use the sub-classes")

    def resample(self, data, **kwargs):  # noqa: D417
        """Resample `data` by calling `precompute` and `compute` methods.

        Args:
            data (xarray.DataArray): Data to be resampled

        Returns (xarray.DataArray): Data resampled to the target area

        """
        self.precompute(**kwargs)
        attrs = data.attrs.copy()
        data_arr = data.data
        dims = _get_dims(data)
        LOG.debug("Resampling %s", str(data.attrs.get("_satpy_id", "unknown")))
        result = self.compute(data_arr, **kwargs)
        coords, result, dims = _check_coords_results_dims(result, data, dims)

        self._adjust_attrs(attrs)

        result = xr.DataArray(result, dims=dims, coords=coords,
                              attrs=attrs)

        return _update_resampled_coords(data, result, self.target_geo_def)

    def _adjust_attrs(self, attrs):
        # Adjust some attributes
        if "BucketFraction" in str(self):
            attrs["units"] = ""
            attrs["calibration"] = ""
            attrs["standard_name"] = "area_fraction"
        elif "BucketCount" in str(self):
            attrs["units"] = ""
            attrs["calibration"] = ""
            attrs["standard_name"] = "number_of_observations"


def _get_dims(data):
    if data.ndim == 3 and data.dims[0] == "bands":
        dims = ("bands", "y", "x")
    # Both one and two dimensional input data results in 2D output
    elif data.ndim in (1, 2):
        dims = ("y", "x")
    else:
        dims = data.dims
    return dims


def _check_coords_results_dims(result, data, dims):
    coords = {}
    if "bands" in data.coords:
        coords["bands"] = data.coords["bands"]
    # Fractions are returned in a dict
    elif isinstance(result, dict):
        coords["categories"] = sorted(result.keys())
        dims = ("categories", "y", "x")
        new_result = []
        for cat in coords["categories"]:
            new_result.append(result[cat])
        result = da.stack(new_result)
    if result.ndim > len(dims):
        result = da.squeeze(result)

    return coords, result, dims


class BucketAvg(BucketResamplerBase):
    """Class for averaging bucket resampling.

    Bucket resampling calculates the average of all the values that
    are closest to each bin and inside the target area.

    Parameters
    ----------
    fill_value : (float) default: `np.nan`
        Fill value to mark missing/invalid values in the input data,
        as well as in the binned and averaged output data.
    skipna : (bool) default: `True`
        If True, skips missing values (as marked by NaN or `fill_value`) for the average calculation
        (similarly to Numpy's `nanmean`). Buckets containing only missing values are set to fill_value.
        If False, sets the bucket to fill_value if one or more missing values are present in the bucket
        (similarly to Numpy's `mean`).
        In both cases, empty buckets are set to `fill_value`.

    """

    def compute(self, data, fill_value=np.nan, skipna=True, **kwargs):  # noqa: D417
        """Call the resampling.

        Args:
            data (numpy.ndarray | dask.array.Array): Data to be resampled
            fill_value (float | int): fill_value. Defaults to numpy.nan
            skipna (bool): Skip NA's. Default `True`

        Returns:
            dask.array.Array
        """
        results = []
        if data.ndim == 3:
            for i in range(data.shape[0]):
                res = self.resampler.get_average(data[i, :, :],
                                                 fill_value=fill_value,
                                                 skipna=skipna,
                                                 **kwargs)
                results.append(res)
        else:
            res = self.resampler.get_average(data, fill_value=fill_value, skipna=skipna,
                                             **kwargs)
            results.append(res)

        return da.stack(results)


class BucketSum(BucketResamplerBase):
    """Class for bucket resampling which implements accumulation (sum).

    This resampler calculates the cumulative sum of all the values
    that are closest to each bin and inside the target area.

    Parameters
    ----------
    fill_value : (float) default: `np.nan`
        Fill value for missing data
    skipna : (bool) default: `True`
        If True, skips NaN values for the sum calculation
        (similarly to Numpy's `nansum`). Buckets containing only NaN are set to zero.
        If False, sets the bucket to NaN if one or more NaN values are present in the bucket
        (similarly to Numpy's `sum`).
        In both cases, empty buckets are set to 0.

    """

    def compute(self, data, skipna=True, **kwargs):
        """Call the resampling."""
        results = []
        if data.ndim == 3:
            for i in range(data.shape[0]):
                res = self.resampler.get_sum(data[i, :, :], skipna=skipna,
                                             **kwargs)
                results.append(res)
        else:
            res = self.resampler.get_sum(data, skipna=skipna, **kwargs)
            results.append(res)

        return da.stack(results)


class BucketCount(BucketResamplerBase):
    """Class for bucket resampling which implements hit-counting.

    This resampler calculates the number of occurences of the input
    data closest to each bin and inside the target area.

    """

    def compute(self, data, **kwargs):
        """Call the resampling."""
        results = []
        if data.ndim == 3:
            for _i in range(data.shape[0]):
                res = self.resampler.get_count()
                results.append(res)
        else:
            res = self.resampler.get_count()
            results.append(res)

        return da.stack(results)


class BucketFraction(BucketResamplerBase):
    """Class for bucket resampling to compute category fractions.

    This resampler calculates the fraction of occurences of the input
    data per category.

    """

    def compute(self, data, fill_value=np.nan, categories=None, **kwargs):
        """Call the resampling."""
        if data.ndim > 2:
            raise ValueError("BucketFraction not implemented for 3D datasets")

        result = self.resampler.get_fractions(data, categories=categories,
                                              fill_value=fill_value)

        return result


def get_resampler_classes():
    """Get bucket resampler classes."""
    return {
        "bucket_avg": BucketAvg,
        "bucket_sum": BucketSum,
        "bucket_count": BucketCount,
        "bucket_fraction": BucketFraction
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""SEVIRI netcdf format reader."""

import datetime as dt
import logging

import numpy as np

from satpy._compat import cached_property
from satpy.readers.core._geos_area import get_area_definition, get_geos_area_naming
from satpy.readers.core.eum import get_service_mode
from satpy.readers.core.file_handlers import BaseFileHandler, open_dataset
from satpy.readers.core.seviri import (
    CHANNEL_NAMES,
    SATNUM,
    CalibParams,
    NominalCoefficients,
    NoValidOrbitParams,
    OrbitPolynomialFinder,
    ScanParams,
    SEVIRICalibrationHandler,
    add_scanline_acq_time,
    create_coef_dict,
    get_cds_time,
    get_satpos,
    mask_bad_quality,
    round_nom_time,
)
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger("nc_msg")
CHUNK_SIZE = get_legacy_chunk_size()


class NCSEVIRIFileHandler(BaseFileHandler):
    """File handler for NC seviri files.

    **Calibration**

    See :mod:`satpy.readers.core.seviri`. Note that there is only one set of
    calibration coefficients available in the netCDF files and therefore there
    is no `calib_mode` argument.

    **Metadata**

    See :mod:`satpy.readers.core.seviri`.

    """

    def __init__(self, filename, filename_info, filetype_info,
                 ext_calib_coefs=None, mask_bad_quality_scan_lines=True):
        """Init the file handler."""
        super(NCSEVIRIFileHandler, self).__init__(filename, filename_info, filetype_info)
        self.ext_calib_coefs = ext_calib_coefs or {}
        self.mask_bad_quality_scan_lines = mask_bad_quality_scan_lines
        self.mda = {}
        self.reference = dt.datetime(1958, 1, 1)
        self.get_metadata()

    @property
    def _repeat_cycle_duration(self):
        """Get repeat cycle duration from the metadata."""
        if self.nc.attrs["nominal_image_scanning"] == "T":
            return 15
        elif self.nc.attrs["reduced_scanning"] == "T":
            return 5

    @property
    def nominal_start_time(self):
        """Read the repeat cycle nominal start time from metadata and round it to expected nominal time slot."""
        tm = self.deltaSt
        return round_nom_time(tm, time_delta=dt.timedelta(minutes=self._repeat_cycle_duration))

    @property
    def nominal_end_time(self):
        """Read the repeat cycle nominal end time from metadata and round it to expected nominal time slot."""
        tm = self.deltaEnd
        return round_nom_time(tm, time_delta=dt.timedelta(minutes=self._repeat_cycle_duration))

    @property
    def observation_start_time(self):
        """Get the repeat cycle observation start time from metadata."""
        return self.deltaSt

    @property
    def observation_end_time(self):
        """Get the repeat cycle observation end time from metadata."""
        return self.deltaEnd

    @property
    def start_time(self):
        """Get general start time for this file."""
        return self.nominal_start_time

    @property
    def end_time(self):
        """Get the general end time for this file."""
        return self.nominal_end_time

    @cached_property
    def nc(self):
        """Read the file."""
        return open_dataset(self.filename, decode_cf=True, mask_and_scale=False,
                            chunks=CHUNK_SIZE).rename({"num_columns_vis_ir": "x",
                                                      "num_rows_vis_ir": "y"})

    def get_metadata(self):
        """Get metadata."""
        # Obtain some area definition attributes
        equatorial_radius = self.nc.attrs["equatorial_radius"] * 1000.
        polar_radius = (self.nc.attrs["north_polar_radius"] * 1000 + self.nc.attrs["south_polar_radius"] * 1000) * 0.5
        ssp_lon = self.nc.attrs["longitude_of_SSP"]
        self.mda["vis_ir_grid_origin"] = self.nc.attrs["vis_ir_grid_origin"]
        self.mda["vis_ir_column_dir_grid_step"] = self.nc.attrs["vis_ir_column_dir_grid_step"] * 1000.0
        self.mda["vis_ir_line_dir_grid_step"] = self.nc.attrs["vis_ir_line_dir_grid_step"] * 1000.0
        # if FSFile is used h5netcdf engine is used which outputs arrays instead of floats for attributes
        if isinstance(equatorial_radius, np.ndarray):
            equatorial_radius = equatorial_radius.item()
            polar_radius = polar_radius.item()
            ssp_lon = ssp_lon.item()
            self.mda["vis_ir_column_dir_grid_step"] = self.mda["vis_ir_column_dir_grid_step"].item()
            self.mda["vis_ir_line_dir_grid_step"] = self.mda["vis_ir_line_dir_grid_step"].item()

        self.mda["projection_parameters"] = {"a": equatorial_radius,
                                             "b": polar_radius,
                                             "h": 35785831.00,
                                             "ssp_longitude": ssp_lon}

        self.mda["number_of_lines"] = int(self.nc.sizes["y"])
        self.mda["number_of_columns"] = int(self.nc.sizes["x"])

        # only needed for HRV channel which is not implemented yet
        # self.mda['hrv_number_of_lines'] = int(self.nc.dims['num_rows_hrv'])
        # self.mda['hrv_number_of_columns'] = int(self.nc.dims['num_columns_hrv'])

        self.deltaSt = self.reference + dt.timedelta(
            days=int(self.nc.attrs["true_repeat_cycle_start_day"]),
            milliseconds=int(self.nc.attrs["true_repeat_cycle_start_mi_sec"]))

        self.deltaEnd = self.reference + dt.timedelta(
            days=int(self.nc.attrs["planned_repeat_cycle_end_day"]),
            milliseconds=int(self.nc.attrs["planned_repeat_cycle_end_mi_sec"]))

        self.north = int(self.nc.attrs["north_most_line"])
        self.east = int(self.nc.attrs["east_most_pixel"])
        self.west = int(self.nc.attrs["west_most_pixel"])
        self.south = int(self.nc.attrs["south_most_line"])
        self.platform_id = int(self.nc.attrs["satellite_id"])

    def get_dataset(self, dataset_id, dataset_info):
        """Get the dataset."""
        dataset = self.nc[dataset_info["nc_key"]]

        # Correct for the scan line order
        # TODO: Move _add_scanline_acq_time() call to the end of the method
        #       once flipping is removed.
        self._add_scanline_acq_time(dataset, dataset_id)
        dataset = dataset.sel(y=slice(None, None, -1))

        dataset = self.calibrate(dataset, dataset_id)
        is_calibration = dataset_id["calibration"] in ["radiance", "reflectance", "brightness_temperature"]
        if (is_calibration and self.mask_bad_quality_scan_lines):
            dataset = self._mask_bad_quality(dataset, dataset_info)

        self._update_attrs(dataset, dataset_info)
        return dataset

    def calibrate(self, dataset, dataset_id):
        """Calibrate the data."""
        channel = dataset_id["name"]
        calibration = dataset_id["calibration"]

        if dataset_id["calibration"] == "counts":
            dataset.attrs["_FillValue"] = 0

        calib = self._get_calibration_handler(dataset, channel)
        return calib.calibrate(dataset, calibration)

    def _get_calibration_handler(self, dataset, channel):
        calib_params = CalibParams(
            mode="NOMINAL",
            internal_coefs=self._get_calib_coefs(dataset, channel),
            external_coefs=self.ext_calib_coefs,
            radiance_type=self._get_radiance_type(channel)
        )
        scan_params = ScanParams(
            int(self.platform_id),
            channel,
            self.observation_start_time
        )
        return SEVIRICalibrationHandler(calib_params, scan_params)

    def _get_radiance_type(self, channel):
        band_idx = list(CHANNEL_NAMES.values()).index(channel)
        return self.nc["planned_chan_processing"].values[band_idx]

    def _get_calib_coefs(self, dataset, channel):
        """Get coefficients for calibration from counts to radiance.

        Only nominal calibration coefficients are available in netCDF files.
        """
        offset = dataset.attrs["add_offset"].astype("float32")
        gain = dataset.attrs["scale_factor"].astype("float32")
        nominal_coefs = NominalCoefficients(channel, gain, offset)
        return create_coef_dict(nominal_coefs)

    def _mask_bad_quality(self, dataset, dataset_info):
        """Mask scanlines with bad quality."""
        ch_number = int(dataset_info["nc_key"][2:])
        line_validity = self.nc["channel_data_visir_data_line_validity"][:, ch_number - 1].data
        line_geometric_quality = self.nc["channel_data_visir_data_line_geometric_quality"][:, ch_number - 1].data
        line_radiometric_quality = self.nc["channel_data_visir_data_line_radiometric_quality"][:, ch_number - 1].data
        return mask_bad_quality(dataset, line_validity, line_geometric_quality, line_radiometric_quality)

    def _update_attrs(self, dataset, dataset_info):
        """Update dataset attributes."""
        dataset.attrs.update(self.nc[dataset_info["nc_key"]].attrs)
        dataset.attrs.update(dataset_info)
        dataset.attrs["platform_name"] = "Meteosat-" + SATNUM[self.platform_id]
        dataset.attrs["sensor"] = "seviri"
        dataset.attrs["orbital_parameters"] = {
            "projection_longitude": self.mda["projection_parameters"]["ssp_longitude"],
            "projection_latitude": 0.,
            "projection_altitude": self.mda["projection_parameters"]["h"],
            "satellite_nominal_longitude": float(
                self.nc.attrs["nominal_longitude"]
            ),
            "satellite_nominal_latitude": 0.0,
        }
        dataset.attrs["time_parameters"] = {
            "nominal_start_time": self.nominal_start_time,
            "nominal_end_time": self.nominal_end_time,
            "observation_start_time": self.observation_start_time,
            "observation_end_time": self.observation_end_time,
        }
        try:
            actual_lon, actual_lat, actual_alt = self.satpos
            dataset.attrs["orbital_parameters"].update({
                "satellite_actual_longitude": actual_lon,
                "satellite_actual_latitude": actual_lat,
                "satellite_actual_altitude": actual_alt,
            })
        except NoValidOrbitParams as err:
            logger.warning(err)
        dataset.attrs["georef_offset_corrected"] = self._get_earth_model() == 2

        # remove attributes from original file which don't apply anymore
        strip_attrs = ["comment", "long_name", "nc_key", "scale_factor", "add_offset", "valid_min", "valid_max"]
        for a in strip_attrs:
            dataset.attrs.pop(a)

    def get_area_def(self, dataset_id):
        """Get the area def.

        Note that the AreaDefinition area extents returned by this function for NetCDF data will be slightly
        different compared to the area extents returned by the SEVIRI HRIT reader.
        This is due to slightly different pixel size values when calculated using the data available in the files. E.g.
        for the 3 km grid:

        ``NetCDF:  self.nc.attrs['vis_ir_column_dir_grid_step'] == 3000.4031658172607``
        ``HRIT: np.deg2rad(2.**16 / pdict['lfac']) * pdict['h'] == 3000.4032785810186``

        This results in the Native 3 km full-disk area extents being approx. 20 cm shorter in each direction.

        The method for calculating the area extents used by the HRIT reader (CFAC/LFAC mechanism) keeps the
        highest level of numeric precision and is used as reference by EUM. For this reason, the standard area
        definitions defined in the `areas.yaml` file correspond to the HRIT ones.

        """
        pdict = {}
        pdict["a"] = self.mda["projection_parameters"]["a"]
        pdict["b"] = self.mda["projection_parameters"]["b"]
        pdict["h"] = self.mda["projection_parameters"]["h"]
        pdict["ssp_lon"] = self.mda["projection_parameters"]["ssp_longitude"]

        area_naming_input_dict = {"platform_name": "msg",
                                  "instrument_name": "seviri",
                                  "resolution": int(dataset_id["resolution"])
                                  }
        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode("seviri", pdict["ssp_lon"])})

        if dataset_id["name"] == "HRV":
            pdict["nlines"] = self.mda["hrv_number_of_lines"]
            pdict["ncols"] = self.mda["hrv_number_of_columns"]
            pdict["a_name"] = area_naming["area_id"]
            pdict["a_desc"] = area_naming["description"]
            pdict["p_id"] = ""
        else:
            pdict["nlines"] = self.mda["number_of_lines"]
            pdict["ncols"] = self.mda["number_of_columns"]
            pdict["a_name"] = area_naming["area_id"]
            pdict["a_desc"] = area_naming["description"]
            pdict["p_id"] = ""

        area = get_area_definition(pdict, self.get_area_extent(dataset_id))

        return area

    def get_area_extent(self, dsid):
        """Get the area extent."""
        # following calculations assume grid origin is south-east corner
        # section 7.2.4 of MSG Level 1.5 Image Data Format Description
        origins = {0: "NW", 1: "SW", 2: "SE", 3: "NE"}
        grid_origin = self.mda["vis_ir_grid_origin"]
        grid_origin = int(grid_origin, 16)
        if grid_origin != 2:
            raise NotImplementedError(
                "Grid origin not supported number: {}, {} corner"
                .format(grid_origin, origins[grid_origin])
            )

        center_point = 3712 / 2

        column_step = self.mda["vis_ir_column_dir_grid_step"]

        line_step = self.mda["vis_ir_line_dir_grid_step"]

        # check for Earth model as this affects the north-south and
        # west-east offsets
        # section 3.1.4.2 of MSG Level 1.5 Image Data Format Description
        earth_model = self._get_earth_model()
        if earth_model == 2:
            ns_offset = 0  # north +ve
            we_offset = 0  # west +ve
        elif earth_model == 1:
            ns_offset = -0.5  # north +ve
            we_offset = 0.5  # west +ve
        else:
            raise NotImplementedError(
                "unrecognised earth model: {}".format(earth_model)
            )
        # section 3.1.5 of MSG Level 1.5 Image Data Format Description
        ll_c = (center_point - self.west - 0.5 + we_offset) * column_step
        ll_l = (self.south - center_point - 0.5 + ns_offset) * line_step
        ur_c = (center_point - self.east + 0.5 + we_offset) * column_step
        ur_l = (self.north - center_point + 0.5 + ns_offset) * line_step
        area_extent = (ll_c, ll_l, ur_c, ur_l)

        return area_extent

    def _add_scanline_acq_time(self, dataset, dataset_id):
        if dataset_id["name"] == "HRV":
            # TODO: Enable once HRV reading has been fixed.
            return
            # days, msecs = self._get_acq_time_hrv()
        else:
            days, msecs = self._get_acq_time_visir(dataset_id)
        acq_time = get_cds_time(days.values, msecs.values)
        add_scanline_acq_time(dataset, acq_time)

    def _get_acq_time_hrv(self):
        day_key = "channel_data_hrv_data_l10_line_mean_acquisition_time_day"
        msec_key = "channel_data_hrv_data_l10_line_mean_acquisition_msec"
        days = self.nc[day_key].isel(channels_hrv_dim=0)
        msecs = self.nc[msec_key].isel(channels_hrv_dim=0)
        return days, msecs

    def _get_acq_time_visir(self, dataset_id):
        band_idx = list(CHANNEL_NAMES.values()).index(dataset_id["name"])
        day_key = "channel_data_visir_data_l10_line_mean_acquisition_time_day"
        msec_key = "channel_data_visir_data_l10_line_mean_acquisition_msec"
        days = self.nc[day_key].isel(channels_vis_ir_dim=band_idx)
        msecs = self.nc[msec_key].isel(channels_vis_ir_dim=band_idx)
        return days, msecs

    @cached_property
    def satpos(self):
        """Get actual satellite position in geodetic coordinates (WGS-84).

        Evaluate orbit polynomials at the start time of the scan.

        Returns: Longitude [deg east], Latitude [deg north] and Altitude [m]
        """
        start_times_poly = get_cds_time(
            days=self.nc["orbit_polynomial_start_time_day"].values,
            msecs=self.nc["orbit_polynomial_start_time_msec"].values
        )
        end_times_poly = get_cds_time(
            days=self.nc["orbit_polynomial_end_time_day"].values,
            msecs=self.nc["orbit_polynomial_end_time_msec"].values
        )
        orbit_polynomials = {
            "StartTime": np.array([start_times_poly]),
            "EndTime": np.array([end_times_poly]),
            "X": self.nc["orbit_polynomial_x"].values,
            "Y": self.nc["orbit_polynomial_y"].values,
            "Z": self.nc["orbit_polynomial_z"].values,
        }
        poly_finder = OrbitPolynomialFinder(orbit_polynomials)
        orbit_polynomial = poly_finder.get_orbit_polynomial(self.start_time)
        return get_satpos(
            orbit_polynomial=orbit_polynomial,
            time=self.start_time,
            semi_major_axis=self.mda["projection_parameters"]["a"],
            semi_minor_axis=self.mda["projection_parameters"]["b"],
        )

    def _get_earth_model(self):
        return int(self.nc.attrs["type_of_earth_model"], 16)


class NCSEVIRIHRVFileHandler(NCSEVIRIFileHandler, SEVIRICalibrationHandler):
    """HRV filehandler."""

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset from file."""
        return NotImplementedError("Currently the HRV channel is not implemented.")

    def get_area_extent(self, dsid):
        """Get HRV area extent."""
        return NotImplementedError



# -*- coding: utf-8 -*-
# Copyright (c) 2017 - 2022 Satpy developers
#
# This file is part of Satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for the GHRSST level-2 formatted data."""

import datetime as dt
import os
import tarfile
from contextlib import suppress
from functools import cached_property

import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()


class GHRSSTL2FileHandler(BaseFileHandler):
    """File handler for GHRSST L2 netCDF files."""

    def __init__(self, filename, filename_info, filetype_info, engine=None):
        """Initialize the file handler for GHRSST L2 netCDF data."""
        super().__init__(filename, filename_info, filetype_info)
        self._engine = engine
        self._tarfile = None

        self.filename_info["start_time"] = dt.datetime.strptime(
            self.nc.start_time, "%Y%m%dT%H%M%SZ")
        self.filename_info["end_time"] = dt.datetime.strptime(
            self.nc.stop_time, "%Y%m%dT%H%M%SZ")

    @cached_property
    def nc(self):
        """Get the xarray Dataset for the filename."""
        if os.fspath(self.filename).endswith("tar"):
            file_obj = self._open_tarfile()
        else:
            file_obj = self.filename

        nc = xr.open_dataset(file_obj,
                             decode_cf=True,
                             mask_and_scale=True,
                             engine=self._engine,
                             chunks={"ni": CHUNK_SIZE,
                                     "nj": CHUNK_SIZE})

        return nc.rename({"ni": "x", "nj": "y"})

    def _open_tarfile(self):
        self._tarfile = tarfile.open(name=self.filename, mode="r")
        sst_filename = next((name for name in self._tarfile.getnames()
                             if self._is_sst_file(name)))
        file_obj = self._tarfile.extractfile(sst_filename)
        return file_obj

    @staticmethod
    def _is_sst_file(name):
        """Check if file in the tar archive is a valid SST file."""
        return name.endswith("nc") and "GHRSST-SSTskin" in name

    def get_dataset(self, key, info):
        """Get any available dataset."""
        stdname = info.get("standard_name")
        return self.nc[stdname].squeeze()

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get end time."""
        return self.filename_info["end_time"]

    @property
    def sensor(self):
        """Get the sensor name."""
        return self.nc.attrs["sensor"].lower()

    def __del__(self):
        """Close the tarfile object."""
        with suppress(AttributeError):
            self._tarfile.close()



# Copyright (c) 2010-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for Scene conversion functionality."""

import datetime as dt
from datetime import datetime

import numpy as np
import pytest
import xarray as xr
from dask import array as da

from satpy import Scene
from satpy.tests.utils import skip_numba_unstable_if_missing

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - include_test_etc

skip_unstable_numba = pytest.mark.skipif(skip_numba_unstable_if_missing(),
                                         reason="Numba is not compatible with unstable NumPy: {err!s}")


@pytest.mark.usefixtures("include_test_etc")
class TestSceneSerialization:
    """Test the Scene serialization."""

    def test_serialization_with_readers_and_data_arr(self):
        """Test that dask can serialize a Scene with readers."""
        from distributed.protocol import deserialize, serialize

        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds1"])
        cloned_scene = deserialize(*serialize(scene))
        assert scene._readers.keys() == cloned_scene._readers.keys()
        assert scene.all_dataset_ids == scene.all_dataset_ids


class TestSceneConversions:
    """Test Scene conversion to geoviews, xarray, etc."""

    def test_geoviews_basic_with_area(self):
        """Test converting a Scene to geoviews with an AreaDefinition."""
        from pyresample.geometry import AreaDefinition
        scn = Scene()
        area = AreaDefinition("test", "test", "test",
                              {"proj": "geos", "lon_0": -95.5, "h": 35786023.0},
                              2, 2, [-200, -200, 200, 200])
        scn["ds1"] = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=("y", "x"),
                                  attrs={"start_time": dt.datetime(2018, 1, 1),
                                         "area": area})
        gv_obj = scn.to_geoviews()
        # we assume that if we got something back, geoviews can use it
        assert gv_obj is not None

    def test_geoviews_basic_with_swath(self):
        """Test converting a Scene to geoviews with a SwathDefinition."""
        from pyresample.geometry import SwathDefinition
        scn = Scene()
        lons = xr.DataArray(da.zeros((2, 2)))
        lats = xr.DataArray(da.zeros((2, 2)))
        area = SwathDefinition(lons, lats)
        scn["ds1"] = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=("y", "x"),
                                  attrs={"start_time": dt.datetime(2018, 1, 1),
                                         "area": area})
        gv_obj = scn.to_geoviews()
        # we assume that if we got something back, geoviews can use it
        assert gv_obj is not None

    @skip_unstable_numba
    def test_hvplot_basic_with_area(self):
        """Test converting a Scene to hvplot with a AreaDefinition."""
        from pyresample.geometry import AreaDefinition
        scn = Scene()
        area = AreaDefinition("test", "test", "test",
                              {"proj": "geos", "lon_0": -95.5, "h": 35786023.0},
                              2, 2, [-200, -200, 200, 200])
        scn["ds1"] = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=("y", "x"),
                                  attrs={"start_time": dt.datetime(2018, 1, 1),
                                         "area": area, "units": "m"})
        hv_obj = scn.to_hvplot()
        # we assume that if we got something back, hvplot can use it
        assert hv_obj is not None

    @skip_unstable_numba
    def test_hvplot_rgb_with_area(self):
        """Test converting a Scene to hvplot with a AreaDefinition."""
        from pyresample.geometry import AreaDefinition
        scn = Scene()
        area = AreaDefinition("test", "test", "test",
                              {"proj": "geos", "lon_0": -95.5, "h": 35786023.0},
                              2, 2, [-200, -200, 200, 200])
        scn["ds1"] = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=("y", "x"),
                                  attrs={"start_time": dt.datetime(2018, 1, 1),
                                         "area": area, "units": "m"})
        scn["ds2"] = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=("y", "x"),
                                  attrs={"start_time": dt.datetime(2018, 1, 1),
                                         "area": area, "units": "m"})
        scn["ds3"] = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=("y", "x"),
                                  attrs={"start_time": dt.datetime(2018, 1, 1),
                                         "area": area, "units": "m"})
        hv_obj = scn.to_hvplot()
        # we assume that if we got something back, hvplot can use it
        assert hv_obj is not None

    @skip_unstable_numba
    def test_hvplot_basic_with_swath(self):
        """Test converting a Scene to hvplot with a SwathDefinition."""
        from pyresample.geometry import SwathDefinition
        scn = Scene()
        longitude = xr.DataArray(da.zeros((2, 2)))
        latitude = xr.DataArray(da.zeros((2, 2)))
        area = SwathDefinition(longitude, latitude)
        scn["ds1"] = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=("y", "x"),
                                  attrs={"start_time": dt.datetime(2018, 1, 1),
                                         "area": area, "units": "m"})
        hv_obj = scn.to_hvplot()
        # we assume that if we got something back, hvplot can use it
        assert hv_obj is not None

class TestToXarrayConversion:
    """Test Scene.to_xarray() conversion."""

    def test_with_empty_scene(self):
        """Test converting empty Scene to xarray."""
        scn = Scene()
        ds = scn.to_xarray()
        assert isinstance(ds, xr.Dataset)
        assert len(ds.variables) == 0
        assert len(ds.coords) == 0

    @pytest.fixture
    def single_area_scn(self):
        """Define Scene with single area."""
        from pyresample.geometry import AreaDefinition

        area = AreaDefinition("test", "test", "test",
                              {"proj": "geos", "lon_0": -95.5, "h": 35786023.0},
                              2, 2, [-200, -200, 200, 200])
        data_array = xr.DataArray(da.zeros((2, 2), chunks=-1),
                                  dims=("y", "x"),
                                  attrs={"start_time": dt.datetime(2018, 1, 1), "area": area})
        scn = Scene()
        scn["var1"] = data_array
        return scn

    def test_to_xarray_dataset_with_conflicting_variables(self):
        """Test converting Scene with DataArrays with conflicting variables.

        E.g. "acq_time" in the seviri_l1b_nc reader
        """
        from pyresample.geometry import AreaDefinition
        area = AreaDefinition("test", "test", "test",
                              {"proj": "geos", "lon_0": -95.5, "h": 35786023.0},
                              2, 2, [-200, -200, 200, 200])
        scn = Scene()

        acq_time_1 = ("y", [np.datetime64("1958-01-02 00:00:01"),
                            np.datetime64("1958-01-02 00:00:02")])
        ds = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=("y", "x"),
                attrs={"start_time": datetime(2018, 1, 1), "area": area})
        ds["acq_time"] = acq_time_1

        scn["ds1"] = ds

        acq_time_2 = ("y", [np.datetime64("1958-02-02 00:00:01"),
                            np.datetime64("1958-02-02 00:00:02")])
        ds2 = ds.copy()
        ds2["acq_time"] = acq_time_2

        scn["ds2"] = ds2

        # drop case (compat="minimal")
        xrds = scn.to_xarray_dataset()
        assert isinstance(xrds, xr.Dataset)
        assert "acq_time" not in xrds.coords

        # override: pick variable from first dataset
        xrds = scn.to_xarray_dataset(datasets=["ds1", "ds2"], compat="override")
        assert isinstance(xrds, xr.Dataset)
        assert "acq_time" in xrds.coords
        xr.testing.assert_equal(xrds["acq_time"], ds["acq_time"])

    @pytest.fixture
    def multi_area_scn(self):
        """Define Scene with multiple area."""
        from pyresample.geometry import AreaDefinition

        area1 = AreaDefinition("test", "test", "test",
                               {"proj": "geos", "lon_0": -95.5, "h": 35786023.0},
                               2, 2, [-200, -200, 200, 200])
        area2 = AreaDefinition("test", "test", "test",
                               {"proj": "geos", "lon_0": -95.5, "h": 35786023.0},
                               4, 4, [-200, -200, 200, 200])

        data_array1 = xr.DataArray(da.zeros((2, 2), chunks=-1),
                                   dims=("y", "x"),
                                   attrs={"start_time": dt.datetime(2018, 1, 1), "area": area1})
        data_array2 = xr.DataArray(da.zeros((4, 4), chunks=-1),
                                   dims=("y", "x"),
                                   attrs={"start_time": dt.datetime(2018, 1, 1), "area": area2})
        scn = Scene()
        scn["var1"] = data_array1
        scn["var2"] = data_array2
        return scn

    def test_with_single_area_scene_type(self, single_area_scn):
        """Test converting single area Scene to xarray dataset."""
        ds = single_area_scn.to_xarray()
        assert isinstance(ds, xr.Dataset)
        assert "var1" in ds.data_vars

    def test_include_lonlats_true(self, single_area_scn):
        """Test include lonlats."""
        ds = single_area_scn.to_xarray(include_lonlats=True)
        assert "latitude" in ds.coords
        assert "longitude" in ds.coords

    def test_include_lonlats_false(self, single_area_scn):
        """Test exclude lonlats."""
        ds = single_area_scn.to_xarray(include_lonlats=False)
        assert "latitude" not in ds.coords
        assert "longitude" not in ds.coords

    def test_dataset_string_accepted(self, single_area_scn):
        """Test accept dataset string."""
        ds = single_area_scn.to_xarray(datasets="var1")
        assert isinstance(ds, xr.Dataset)

    def test_wrong_dataset_key(self, single_area_scn):
        """Test raise error if unexisting dataset."""
        with pytest.raises(KeyError):
            _ = single_area_scn.to_xarray(datasets="var2")

    def test_to_xarray_with_multiple_area_scene(self, multi_area_scn):
        """Test converting muiltple area Scene to xarray."""
        # TODO: in future adapt for DataTree implementation
        with pytest.raises(ValueError, match="Datasets to be saved .* must have identical projection coordinates."):
            _ = multi_area_scn.to_xarray()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for Himawari L2 cloud products from NOAA's big data programme.

For more information about the data, see: <https://registry.opendata.aws/noaa-himawari/>.

These products are generated by the NOAA enterprise cloud suite and have filenames like:
AHI-CMSK_v1r1_h09_s202308240540213_e202308240549407_c202308240557548.nc

The second letter grouping (CMSK above) indicates the product type:
    CMSK - Cloud mask

    CHGT - Cloud height

    CPHS - Cloud type and phase

These products are generated from the AHI sensor on Himawari-8 and Himawari-9, and are
produced at the native instrument resolution for the IR channels (2km at nadir).

NOTE: This reader is currently only compatible with full disk scenes. Unlike level 1 himawari
data, the netCDF files do not contain the required metadata to produce an appropriate area
definition for the data contents, and hence the area definition is hardcoded into the reader.

A warning is displayed to the user highlighting this. The assumed area definition is a full
disk image at the nominal subsatellite longitude of 140.7 degrees East.

All the simple data products are supported here, but multidimensional products are not yet
supported. These include the CldHgtFlag and the CloudMaskPacked variables.
"""

import datetime as dt
import logging

import xarray as xr

from satpy._compat import cached_property
from satpy.readers.core._geos_area import get_area_definition, get_area_extent
from satpy.readers.core.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)

EXPECTED_DATA_AREA = "Full Disk"


class HIML2NCFileHandler(BaseFileHandler):
    """File handler for Himawari L2 NOAA enterprise data in netCDF format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super().__init__(filename, filename_info, filetype_info)
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  chunks={"xc": "auto", "yc": "auto"})

        # Check that file is a full disk scene, we don't know the area for anything else
        if self.nc.attrs["cdm_data_type"] != EXPECTED_DATA_AREA:
            raise ValueError("File is not a full disk scene")

        self.sensor = self.nc.attrs["instrument_name"].lower()
        self.nlines = self.nc.sizes["Columns"]
        self.ncols = self.nc.sizes["Rows"]
        self.platform_name = self.nc.attrs["satellite_name"]
        self.platform_shortname = filename_info["platform"]
        self._meta = None

    @property
    def start_time(self):
        """Start timestamp of the dataset."""
        date_str = self.nc.attrs["time_coverage_start"]
        return dt.datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ")

    @property
    def end_time(self):
        """End timestamp of the dataset."""
        date_str = self.nc.attrs["time_coverage_end"]
        return dt.datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ")

    def get_dataset(self, key, info):
        """Load a dataset."""
        var = info["file_key"]
        logger.debug("Reading in get_dataset %s.", var)
        variable = self.nc[var]

        # Data has 'Latitude' and 'Longitude' coords, these must be replaced.
        variable = variable.rename({"Rows": "y", "Columns": "x"})

        variable = variable.drop_vars("Latitude")
        variable = variable.drop_vars("Longitude")

        variable.attrs["sensor"] = self.sensor
        variable.attrs["platform_name"] = self.platform_name
        variable.attrs["platform_shortname"] = self.platform_shortname
        variable.attrs["start_time"] = self.start_time
        variable.attrs["end_time"] = self.end_time
        variable.attrs.update(key.to_dict())
        return variable

    @cached_property
    def area(self):
        """Get AreaDefinition representing this file's data."""
        return self._get_area_def()

    def get_area_def(self, dsid):
        """Get the area definition."""
        del dsid
        return self.area

    def _get_area_def(self):
        logger.info("The AHI L2 cloud products do not have the metadata required to produce an area definition."
                    " Assuming standard Himawari-8/9 full disk projection.")

        # Basic check to ensure we're processing a full disk (2km) scene.n
        if self.nlines != 5500 or self.ncols != 5500:
            raise ValueError("Input L2 file is not a full disk Himawari scene. Only full disk data is supported.")

        pdict = {"cfac": 20466275, "lfac": 20466275, "coff": 2750.5, "loff": 2750.5, "a": 6378137.0, "h": 35785863.0,
                 "b": 6356752.3, "ssp_lon": 140.7, "nlines": self.nlines, "ncols": self.ncols, "scandir": "N2S"}

        aex = get_area_extent(pdict)

        pdict["a_name"] = "Himawari_Area"
        pdict["a_desc"] = "AHI Full Disk area"
        pdict["p_id"] = f"geos{self.platform_shortname}"

        return get_area_definition(pdict, aex)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Advanced Meteorological Imager reader for the Level 1b NetCDF4 format."""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import pyproj
import xarray as xr
from pyspectral.blackbody import blackbody_wn_rad2temp as rad2temp

import satpy
from satpy.readers.core._geos_area import get_area_definition, get_area_extent
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.remote import open_file_or_filename
from satpy.readers.core.utils import apply_rad_correction, get_user_calibration_factors
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()
PLATFORM_NAMES = {
    "GK-2A": "GEO-KOMPSAT-2A",
    "GK-2B": "GEO-KOMPSAT-2B",
}


class AMIL1bNetCDF(BaseFileHandler):
    """Base reader for AMI L1B NetCDF4 files.

    AMI data contains GSICS adjustment factors for the IR bands.
    By default, these are not applied. If you wish to apply them then you must
    set the calibration mode appropriately::

        import satpy
        import glob

        filenames = glob.glob('*FLDK*.dat')
        scene = satpy.Scene(filenames,
                            reader='ahi_hsd',
                            reader_kwargs={'calib_mode': 'gsics'})
        scene.load(['B13'])

    In addition, the GSICS website (and other sources) also supply radiance
    correction coefficients like so::

        radiance_corr = (radiance_orig - corr_offset) / corr_slope

    If you wish to supply such coefficients, pass 'user_calibration' and a
    dictionary containing per-channel slopes and offsets as a reader_kwarg::

       user_calibration={'chan': {'slope': slope, 'offset': offset}}

    If you do not have coefficients for a particular band, then by default the
    slope will be set to 1 .and the offset to 0.::

        import satpy
        import glob

        # Load bands 7, 14 and 15, but we only have coefs for 7+14
        calib_dict = {'WV063': {'slope': 0.99, 'offset': 0.002},
                      'IR087': {'slope': 1.02, 'offset': -0.18}}

        filenames = glob.glob('*.nc')
        scene = satpy.Scene(filenames,
                            reader='ami_l1b',
                            reader_kwargs={'user_calibration': calib_dict,
                                           'calib_mode': 'file')
        # IR133 will not have radiance correction applied.
        scene.load(['WV063', 'IR087', 'IR133'])

    By default these updated coefficients are not used. In most cases, setting
    `calib_mode` to `file` is required in order to use external coefficients.
    """

    def __init__(self, filename, filename_info, filetype_info,
                 calib_mode="PYSPECTRAL", allow_conditional_pixels=False,
                 user_calibration=None, clip_negative_radiances=None):
        """Open the NetCDF file with xarray and prepare the Dataset for reading."""
        super(AMIL1bNetCDF, self).__init__(filename, filename_info, filetype_info)
        f_obj = open_file_or_filename(self.filename)
        self.nc = xr.open_dataset(f_obj,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  chunks={"dim_image_x": CHUNK_SIZE, "dim_image_y": CHUNK_SIZE})
        self.nc = self.nc.rename({"dim_image_x": "x", "dim_image_y": "y"})

        platform_shortname = self.nc.attrs["satellite_name"]
        self.platform_name = PLATFORM_NAMES.get(platform_shortname)
        self.sensor = "ami"
        self.band_name = filetype_info["file_type"].upper()
        self.allow_conditional_pixels = allow_conditional_pixels
        calib_mode_choices = ("FILE", "PYSPECTRAL", "GSICS")
        if calib_mode.upper() not in calib_mode_choices:
            raise ValueError("Invalid calibration mode: {}. Choose one of {}".format(
                calib_mode, calib_mode_choices))

        self.calib_mode = calib_mode.upper()
        self.user_calibration = user_calibration

        if clip_negative_radiances is None:
            clip_negative_radiances = satpy.config.get("readers.clip_negative_radiances")
        self.clip_negative_radiances = clip_negative_radiances

    @property
    def start_time(self):
        """Get observation start time."""
        base = dt.datetime(2000, 1, 1, 12, 0, 0)
        return base + dt.timedelta(seconds=self.nc.attrs["observation_start_time"])

    @property
    def end_time(self):
        """Get observation end time."""
        base = dt.datetime(2000, 1, 1, 12, 0, 0)
        return base + dt.timedelta(seconds=self.nc.attrs["observation_end_time"])

    def get_area_def(self, dsid):
        """Get area definition for this file."""
        pdict = {}
        pdict["a"] = self.nc.attrs["earth_equatorial_radius"]
        pdict["b"] = self.nc.attrs["earth_polar_radius"]
        pdict["h"] = self.nc.attrs["nominal_satellite_height"] - pdict["a"]
        pdict["ssp_lon"] = self.nc.attrs["sub_longitude"] * 180 / np.pi  # it's in radians?
        pdict["ncols"] = self.nc.attrs["number_of_columns"]
        pdict["nlines"] = self.nc.attrs["number_of_lines"]
        obs_mode = self.nc.attrs["observation_mode"]
        resolution = self.nc.attrs["channel_spatial_resolution"]

        # Example offset: 11000.5
        # the 'get_area_extent' will handle this half pixel for us
        pdict["cfac"] = self.nc.attrs["cfac"]
        pdict["coff"] = self.nc.attrs["coff"]
        pdict["lfac"] = -self.nc.attrs["lfac"]
        pdict["loff"] = self.nc.attrs["loff"]
        pdict["scandir"] = "N2S"
        pdict["a_name"] = "ami_geos_{}".format(obs_mode.lower())
        pdict["a_desc"] = "AMI {} Area at {} resolution".format(obs_mode, resolution)
        pdict["p_id"] = "ami_fixed_grid"

        area_extent = get_area_extent(pdict)
        fg_area_def = get_area_definition(pdict, area_extent)
        return fg_area_def

    def get_orbital_parameters(self):
        """Collect orbital parameters for this file."""
        a = float(self.nc.attrs["earth_equatorial_radius"])
        b = float(self.nc.attrs["earth_polar_radius"])
        # nominal_satellite_height seems to be from the center of the earth
        h = float(self.nc.attrs["nominal_satellite_height"]) - a
        lon_0 = self.nc.attrs["sub_longitude"] * 180 / np.pi  # it's in radians?
        sc_position = self.nc["sc_position"].attrs["sc_position_center_pixel"]

        # convert ECEF coordinates to lon, lat, alt
        ecef = pyproj.CRS.from_dict({"proj": "geocent", "a": a, "b": b})
        lla = pyproj.CRS.from_dict({"proj": "latlong", "a": a, "b": b})
        transformer = pyproj.Transformer.from_crs(ecef, lla)
        sc_position = transformer.transform(sc_position[0], sc_position[1], sc_position[2])

        orbital_parameters = {
            "projection_longitude": float(lon_0),
            "projection_latitude": 0.0,
            "projection_altitude": h,
            "satellite_actual_longitude": sc_position[0],
            "satellite_actual_latitude": sc_position[1],
            "satellite_actual_altitude": sc_position[2],  # meters
        }
        return orbital_parameters

    def get_dataset(self, dataset_id, ds_info):
        """Load a dataset as a xarray DataArray."""
        file_key = ds_info.get("file_key", dataset_id["name"])
        data = self.nc[file_key]
        # hold on to attributes for later
        attrs = data.attrs
        # highest 2 bits are data quality flags
        # 00=no error
        # 01=available under conditions
        # 10=outside the viewing area
        # 11=Error exists
        if self.allow_conditional_pixels:
            qf = data & 0b1000000000000000
        else:
            qf = data & 0b1100000000000000

        # mask DQF bits
        bits = attrs["number_of_valid_bits_per_pixel"].astype(data.dtype)
        data &= 2**bits - 1
        # only take "no error" pixels as valid
        data = data.where(qf == 0)

        # Calibration values from file, fall back to built-in if unavailable
        gain = self.nc.attrs["DN_to_Radiance_Gain"]
        offset = self.nc.attrs["DN_to_Radiance_Offset"]

        if dataset_id["calibration"] in ("radiance", "reflectance", "brightness_temperature"):
            data = gain * data + offset
            data = self._clip_negative_radiance(data, gain, offset)
            if self.calib_mode == "GSICS":
                data = self._apply_gsics_rad_correction(data)
            elif isinstance(self.user_calibration, dict):
                data = self._apply_user_rad_correction(data)
        if dataset_id["calibration"] == "reflectance":
            # depends on the radiance calibration above
            rad_to_alb = self.nc.attrs["Radiance_to_Albedo_c"]
            if ds_info.get("units") == "%":
                rad_to_alb *= 100
            data = data * rad_to_alb
        elif dataset_id["calibration"] == "brightness_temperature":
            data = self._calibrate_ir(dataset_id, data)
        elif dataset_id["calibration"] not in ("counts", "radiance"):
            raise ValueError("Unknown calibration: '{}'".format(dataset_id["calibration"]))

        for attr_name in ("standard_name", "units"):
            attrs[attr_name] = ds_info[attr_name]
        attrs.update(dataset_id.to_dict())
        attrs["orbital_parameters"] = self.get_orbital_parameters()
        attrs["platform_name"] = self.platform_name
        attrs["sensor"] = self.sensor
        data.attrs = attrs
        return data

    def _clip_negative_radiance(self, data, gain, offset):
        """If requested, clip negative radiance from Rad DataArray."""
        if self.clip_negative_radiances:
            count_zero_rad = - offset / gain
            # We need floor here as the scale factor for AMI is negative (unlike ABI)
            count_pos = np.floor(count_zero_rad)
            min_rad = count_pos * gain + offset
            data = data.clip(min=min_rad)
        return data

    def _calibrate_ir(self, dataset_id, data):
        """Calibrate radiance data to BTs using either pyspectral or in-file coefficients."""
        if self.calib_mode == "PYSPECTRAL":
            # depends on the radiance calibration above
            # Convert um to m^-1 (SI units for pyspectral)
            wn = 1 / (dataset_id["wavelength"][1] / 1e6)
            # Convert cm^-1 (wavenumbers) and (mW/m^2)/(str/cm^-1) (radiance data)
            # to SI units m^-1, mW*m^-3*str^-1.
            bt_data = rad2temp(wn, data.data * 1e-5)
            if isinstance(bt_data, np.ndarray):
                # old versions of pyspectral produce numpy arrays
                bt_data = da.from_array(bt_data, chunks=data.data.chunks)
            # new versions of pyspectral can do dask arrays
            data.data = bt_data
        else:
            # IR coefficients from the file
            # Channel specific
            c0 = self.nc.attrs["Teff_to_Tbb_c0"]
            c1 = self.nc.attrs["Teff_to_Tbb_c1"]
            c2 = self.nc.attrs["Teff_to_Tbb_c2"]

            # These should be fixed, but load anyway
            cval = self.nc.attrs["light_speed"]
            kval = self.nc.attrs["Boltzmann_constant_k"]
            hval = self.nc.attrs["Plank_constant_h"]

            # Compute wavenumber as cm-1
            wn = (10000 / dataset_id["wavelength"][1]) * 100
            # Convert radiance to effective brightness temperature
            e1 = (2 * hval * cval * cval) * np.power(wn, 3)
            e2 = (data.data * 1e-5)
            t_eff = ((hval * cval / kval) * wn) / np.log((e1 / e2) + 1)

            # Now convert to actual brightness temperature
            bt_data = c0 + c1 * t_eff + c2 * t_eff * t_eff
            data.data = bt_data
        return data

    def _apply_gsics_rad_correction(self, data):
        """Retrieve GSICS factors from L1 file and apply to radiance."""
        rad_slope = self.nc["gsics_coeff_slope"][0]
        rad_offset = self.nc["gsics_coeff_intercept"][0]
        data = apply_rad_correction(data, rad_slope, rad_offset)
        return data

    def _apply_user_rad_correction(self, data):
        """Retrieve user-supplied radiance correction and apply."""
        rad_slope, rad_offset = get_user_calibration_factors(self.band_name,
                                                             self.user_calibration)
        data = apply_rad_correction(data, rad_slope, rad_offset)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Setup for SEVIRI HRIT reader tests."""

import datetime as dt
from unittest import mock

import numpy as np

from satpy.readers.seviri_l1b_hrit import HRITMSGFileHandler, HRITMSGPrologueFileHandler
from satpy.tests.reader_tests.test_seviri_base import ORBIT_POLYNOMIALS


def new_get_hd(instance, hdr_info):
    """Generate some metadata."""
    mda = {"spectral_channel_id": 1}
    mda.setdefault("number_of_bits_per_pixel", 10)

    mda["projection_parameters"] = {"a": 6378169.00,
                                             "b": 6356583.80,
                                             "h": 35785831.00,
                                             "SSP_longitude": 0.0}
    mda["orbital_parameters"] = {}
    mda["total_header_length"] = 12
    return mda


def get_new_read_prologue(prologue):
    """Create mocked read_prologue() method."""
    def new_read_prologue(self):
        self.prologue = prologue
    return new_read_prologue


def get_fake_file_handler(observation_start_time, nlines, ncols, projection_longitude=0,
                          orbit_polynomials=ORBIT_POLYNOMIALS):
    """Create a mocked SEVIRI HRIT file handler."""
    import warnings

    prologue = get_fake_prologue(projection_longitude, orbit_polynomials)
    mda = get_fake_mda(nlines=nlines, ncols=ncols, start_time=observation_start_time)
    filename_info = get_fake_filename_info(observation_start_time)
    epilogue = get_fake_epilogue()

    m = mock.mock_open()
    with mock.patch("satpy.readers.seviri_l1b_hrit.np.fromfile") as fromfile, \
            mock.patch("satpy.readers.core.hrit.open", m, create=True) as newopen, \
            mock.patch("satpy.readers.core.utils.generic_open", m, create=True) as genopen, \
            mock.patch("satpy.readers.seviri_l1b_hrit.CHANNEL_NAMES"), \
            mock.patch.object(HRITMSGFileHandler, "_get_hd", new=new_get_hd), \
            mock.patch.object(HRITMSGPrologueFileHandler, "read_prologue",
                              new=get_new_read_prologue(prologue)):

        fromfile.return_value = np.array(
            [(1, 2)],
            dtype=[("total_header_length", int),
                   ("hdr_id", int)]
        )
        newopen.return_value.__enter__.return_value.tell.return_value = 1
        # The size of the return value hereafter was chosen arbitrarily with the expectation
        # that it would return sufficiently many bytes for testing the fake-opening of HRIT
        # files.
        genopen.return_value.__enter__.return_value.read.return_value = bytes([0]*8192)
        prologue = HRITMSGPrologueFileHandler(
            filename="dummy_prologue_filename",
            filename_info=filename_info,
            filetype_info={}
        )
        epilogue = mock.MagicMock(epilogue=epilogue)

        with warnings.catch_warnings():
            # Orbit polynomial has no exact match, so filter the unnecessary warning
            warnings.filterwarnings("ignore", category=UserWarning, message=r"No orbit polynomial valid for")
            reader = HRITMSGFileHandler(
                "filename",
                filename_info,
                {"filetype": "info"},
                prologue,
                epilogue
            )
        reader.mda.update(mda)
        return reader


def get_fake_prologue(projection_longitude, orbit_polynomials):
    """Create a fake HRIT prologue."""
    return {
         "SatelliteStatus": {
             "SatelliteDefinition": {
                 "SatelliteId": 324,
                 "NominalLongitude": -3.5
             },
             "Orbit": {
                 "OrbitPolynomial": orbit_polynomials,
             }
         },
         "GeometricProcessing": {
             "EarthModel": {
                 "TypeOfEarthModel": 2,
                 "EquatorialRadius": 6378.169,
                 "NorthPolarRadius": 6356.5838,
                 "SouthPolarRadius": 6356.5838
             }
         },
         "ImageDescription": {
             "ProjectionDescription": {
                 "LongitudeOfSSP": projection_longitude
             },
             "Level15ImageProduction": {
                 "ImageProcDirection": 1
             }
         },
         "ImageAcquisition": {
            "PlannedAcquisitionTime": {
                "TrueRepeatCycleStart": dt.datetime(2006, 1, 1, 12, 15, 9, 304888),
                "PlannedRepeatCycleEnd": dt.datetime(2006, 1, 1, 12, 30, 0, 0)
            }
         }
    }


def get_fake_epilogue():
    """Create a fake HRIT epilogue."""
    return {
            "ImageProductionStats": {
                "ActualL15CoverageHRV": {
                    "LowerSouthLineActual": 1,
                    "LowerNorthLineActual": 8256,
                    "LowerEastColumnActual": 2877,
                    "LowerWestColumnActual": 8444,
                    "UpperSouthLineActual": 8257,
                    "UpperNorthLineActual": 11136,
                    "UpperEastColumnActual": 1805,
                    "UpperWestColumnActual": 7372
                },
                "ActualScanningSummary": {
                    "ReducedScan": 0,
                    "ForwardScanStart": dt.datetime(2006, 1, 1, 12, 15, 9, 304888),
                    "ForwardScanEnd": dt.datetime(2006, 1, 1, 12, 27, 39, 0)
                }
            }
        }


def get_fake_mda(nlines, ncols, start_time):
    """Create fake metadata."""
    nbits = 10
    tline = get_acq_time_cds(start_time, nlines)
    return {
        "number_of_bits_per_pixel": nbits,
        "number_of_lines": nlines,
        "number_of_columns": ncols,
        "data_field_length": nlines * ncols * nbits,
        "cfac": 5,
        "lfac": 5,
        "coff": 10,
        "loff": 10,
        "image_segment_line_quality": {
            "line_mean_acquisition": tline,
            "line_validity": np.full(nlines, 3),
            "line_radiometric_quality": np.full(nlines, 4),
            "line_geometric_quality": np.full(nlines, 4)
        }
    }


def get_fake_filename_info(start_time):
    """Create fake filename information."""
    return {
        "platform_shortname": "MSG3",
        "start_time": start_time,
        "service": "MSG"
    }


def get_fake_dataset_info():
    """Create fake dataset info."""
    return {
        "units": "units",
        "wavelength": "wavelength",
        "standard_name": "standard_name"
    }


def get_acq_time_cds(start_time, nlines):
    """Get fake scanline acquisition times."""
    days_since_1958 = (start_time - dt.datetime(1958, 1, 1)).days
    tline = np.zeros(
        nlines,
        dtype=[("days", ">u2"), ("milliseconds", ">u4")]
    )
    tline["days"][1:-1] = days_since_1958 * np.ones(nlines - 2)
    offset_second = (start_time - start_time.replace(hour=0, minute=0, second=0, microsecond=0)).total_seconds()*1000
    tline["milliseconds"][1:-1] = np.arange(nlines - 2)+offset_second

    return tline


def get_acq_time_exp(start_time, nlines):
    """Get expected scanline acquisition times."""
    tline_exp = np.zeros(464, dtype="datetime64[ms]")
    tline_exp[0] = np.datetime64("NaT")
    tline_exp[-1] = np.datetime64("NaT")
    tline_exp[1:-1] = np.datetime64(start_time)
    tline_exp[1:-1] += np.arange(nlines - 2).astype("timedelta64[ms]")

    return tline_exp.astype("datetime64[ns]")


def get_attrs_exp(projection_longitude=0.0):
    """Get expected dataset attributes."""
    return {
        "units": "units",
        "wavelength": "wavelength",
        "standard_name": "standard_name",
        "platform_name": "Meteosat-11",
        "sensor": "seviri",
        "orbital_parameters": {"projection_longitude": projection_longitude,
                               "projection_latitude": 0.,
                               "projection_altitude": 35785831.0,
                               "satellite_nominal_longitude": -3.5,
                               "satellite_nominal_latitude": 0.0,
                               "satellite_actual_longitude": -3.55117540817073,
                               "satellite_actual_latitude": -0.5711243456528018,
                               "satellite_actual_altitude": 35783296.150123544},
        "georef_offset_corrected": True,
        "nominal_start_time": dt.datetime(2006, 1, 1, 12, 15),
        "nominal_end_time": dt.datetime(2006, 1, 1, 12, 30),
        "time_parameters": {
            "nominal_start_time": dt.datetime(2006, 1, 1, 12, 15),
            "nominal_end_time": dt.datetime(2006, 1, 1, 12, 30),
            "observation_start_time": dt.datetime(2006, 1, 1, 12, 15, 9, 304888),
            "observation_end_time": dt.datetime(2006, 1, 1, 12, 27, 39, 0)
            }
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Modifier classes dealing with spectral domain changes or corrections."""

import logging

import xarray as xr

from satpy.modifiers import ModifierBase

try:
    from pyspectral.near_infrared_reflectance import Calculator
except ImportError:
    Calculator = None

try:
    from pyorbital.astronomy import sun_zenith_angle
except ImportError:
    sun_zenith_angle = None


logger = logging.getLogger(__name__)


class NIRReflectance(ModifierBase):
    """Get the reflective part of NIR bands."""

    TERMINATOR_LIMIT = 85.0
    MASKING_LIMIT = 88.0

    def __init__(self, sunz_threshold=TERMINATOR_LIMIT,  # noqa: D417
                 masking_limit=MASKING_LIMIT, **kwargs):
        """Collect custom configuration values.

        Args:
            sunz_threshold: The threshold sun zenith angle used when deriving
                the near infrared reflectance. Above this angle the derivation
                will assume this sun-zenith everywhere. Unless overridden, the
                default threshold of 85.0 degrees will be used.
            masking_limit: Mask the data (set to NaN) above this Sun zenith angle.
                By default the limit is at 88.0 degrees.  If set to `None`, no masking
                is done.

        """
        self.sun_zenith_threshold = sunz_threshold
        self.masking_limit = masking_limit
        super(NIRReflectance, self).__init__(**kwargs)

    def __call__(self, projectables, optional_datasets=None, **info):
        """Get the reflectance part of an NIR channel.

        Not supposed to be used for wavelength outside [3, 4] µm.
        """
        matched_data_arrs = self.match_data_arrays(projectables + optional_datasets)
        projectables = matched_data_arrs[:len(projectables)]
        optional_datasets = matched_data_arrs[len(projectables):]
        inputs = self._get_nir_inputs(projectables, optional_datasets)
        return self._get_reflectance_as_dataarray(*inputs)

    def _get_reflectance_as_dataarray(self, nir, da_tb11, da_tb13_4, da_sun_zenith):
        """Get the reflectance as a dataarray."""
        logger.info("Getting reflective part of %s", nir.attrs["name"])
        reflectance = self._get_reflectance_as_dask(nir.data, da_tb11, da_tb13_4, da_sun_zenith, nir.attrs)
        proj = self._create_modified_dataarray(reflectance, base_dataarray=nir)
        proj.attrs["units"] = "%"
        return proj

    def _get_nir_inputs(self, projectables, optional_datasets):
        nir, tb11 = projectables
        da_tb11 = tb11.data
        da_tb13_4 = self._get_tb13_4_from_optionals(optional_datasets)
        da_sun_zenith = self._get_sun_zenith_from_provided_data(nir, optional_datasets, nir.dtype)
        return (nir, da_tb11, da_tb13_4, da_sun_zenith)

    @staticmethod
    def _get_tb13_4_from_optionals(optional_datasets):
        tb13_4 = None
        for dataset in optional_datasets:
            wavelengths = dataset.attrs.get("wavelength", [100., 0, 0])
            if (dataset.attrs.get("units") == "K" and
                    wavelengths[0] <= 13.4 <= wavelengths[2]):
                tb13_4 = dataset.data
        return tb13_4

    @staticmethod
    def _get_sun_zenith_from_provided_data(nir, optional_datasets, dtype):
        """Get the sunz from available data or compute it if unavailable."""
        sun_zenith = None

        for dataset in optional_datasets:
            if dataset.attrs.get("standard_name") == "solar_zenith_angle":
                sun_zenith = dataset.data

        if sun_zenith is None:
            if sun_zenith_angle is None:
                raise ImportError("Module pyorbital.astronomy needed to compute sun zenith angles.")
            lons, lats = nir.attrs["area"].get_lonlats(chunks=nir.data.chunks, dtype=dtype)
            sun_zenith = sun_zenith_angle(nir.attrs["start_time"], lons, lats)
        return sun_zenith

    def _create_modified_dataarray(self, reflectance, base_dataarray):
        proj = xr.DataArray(reflectance, dims=base_dataarray.dims,
                            coords=base_dataarray.coords, attrs=base_dataarray.attrs.copy())
        proj.attrs["sun_zenith_threshold"] = self.sun_zenith_threshold
        proj.attrs["sun_zenith_masking_limit"] = self.masking_limit
        self.apply_modifier_info(base_dataarray, proj)
        return proj

    def _get_reflectance_as_dask(self, da_nir, da_tb11, da_tb13_4, da_sun_zenith, metadata):
        """Calculate 3.x reflectance in % with pyspectral from dask arrays."""
        reflectance_3x_calculator = self._init_reflectance_calculator(metadata)
        return reflectance_3x_calculator.reflectance_from_tbs(da_sun_zenith, da_nir, da_tb11, tb_ir_co2=da_tb13_4) * 100

    def _init_reflectance_calculator(self, metadata):
        """Initialize the 3.x reflectance derivations."""
        if not Calculator:
            logger.info("Couldn't load pyspectral")
            raise ImportError("No module named pyspectral.near_infrared_reflectance")

        reflectance_3x_calculator = Calculator(metadata["platform_name"], metadata["sensor"], metadata["name"],
                                               sunz_threshold=self.sun_zenith_threshold,
                                               masking_limit=self.masking_limit)
        return reflectance_3x_calculator


class NIREmissivePartFromReflectance(NIRReflectance):
    """Get the emissive part of NIR bands."""

    def __init__(self, sunz_threshold=None, **kwargs):  # noqa: D417
        """Collect custom configuration values.

        Args:
            sunz_threshold: The threshold sun zenith angle used when deriving
                the near infrared reflectance. Above this angle the derivation
                will assume this sun-zenith everywhere. Default None, in which
                case the default threshold defined in Pyspectral will be used.

        """
        self.sunz_threshold = sunz_threshold
        super(NIREmissivePartFromReflectance, self).__init__(sunz_threshold=sunz_threshold, **kwargs)

    def __call__(self, projectables, optional_datasets=None, **info):
        """Get the emissive part an NIR channel after having derived the reflectance.

        Not supposed to be used for wavelength outside [3, 4] µm.

        """
        projectables = self.match_data_arrays(projectables)
        inputs = self._get_nir_inputs(projectables, optional_datasets)
        return self._get_emissivity_as_dataarray(*inputs)

    def _get_emissivity_as_dataarray(self, nir, da_tb11, da_tb13_4, da_sun_zenith):
        """Get the emissivity as a dataarray."""
        logger.info("Getting emissive part of %s", nir.attrs["name"])
        emissivity = self._get_emissivity_as_dask(nir.data, da_tb11, da_tb13_4, da_sun_zenith, nir.attrs)

        proj = self._create_modified_dataarray(emissivity, base_dataarray=nir)
        proj.attrs["units"] = "K"
        return proj

    def _get_emissivity_as_dask(self, da_nir, da_tb11, da_tb13_4, da_sun_zenith, metadata):
        """Get the emissivity from pyspectral."""
        reflectance_3x_calculator = self._init_reflectance_calculator(metadata)
        # Use the nir and thermal ir brightness temperatures and derive the reflectance using
        # PySpectral. The reflectance is stored internally in PySpectral and
        # needs to be derived first in order to get the emissive part.
        reflectance_3x_calculator.reflectance_from_tbs(da_sun_zenith, da_nir, da_tb11, tb_ir_co2=da_tb13_4)
        return reflectance_3x_calculator.emissive_part_3x()



#!/usr/bin/env python
# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helper functions for remote reading."""

import os
import pathlib
import pickle  # nosec B403
from functools import total_ordering

import fsspec
from upath import UPath


@total_ordering
class FSFile(os.PathLike):
    """Implementation of a PathLike file object, that can be opened.

    Giving the filenames to :class:`Scene <satpy.scene.Scene>` with valid transfer protocols will automatically
    use this class so manual usage of this class is needed mainly for fine-grained control.

    This class is made to be used in conjuction with fsspec or s3fs. For example::

        from satpy import Scene

        import fsspec
        filename = 'noaa-goes16/ABI-L1b-RadC/2019/001/17/*_G16_s20190011702186*'

        the_files = fsspec.open_files("simplecache::s3://" + filename, s3={'anon': True})

        from satpy.readers.core.remote import FSFile
        fs_files = [FSFile(open_file) for open_file in the_files]

        scn = Scene(filenames=fs_files, reader='abi_l1b')
        scn.load(['true_color_raw'])

    """

    def __init__(
            self,
            file: os.PathLike | fsspec.core.OpenFile | str,
            fs: fsspec.spec.AbstractFileSystem | None = None,
    ):
        """Initialise the FSFile instance.

        Args:
            file:
                String, object implementing the :class:`os.PathLike` protocol, or
                an :class:`~fsspec.core.OpenFile` instance.  If passed an instance of
                :class:`~fsspec.core.OpenFile`, the following argument ``fs`` has no
                effect.
            fs:
                Object implementing the fsspec filesystem protocol.
        """
        self._fs_open_kwargs = _get_fs_open_kwargs(file)
        if hasattr(file, "path") and hasattr(file, "fs"):
            self._file = file.path
            self._fs = file.fs
        else:
            self._file = file
            self._fs = fs

    def to_upath(self):
        """Convert to UPath."""
        try:
            kwargs = self._fs.to_dict()
            kwargs.pop("cls")
        except AttributeError:
            kwargs = dict()
        return UPath(self._file, **kwargs)

    def __str__(self):
        """Return the string version of the filename."""
        return os.fspath(self._file)

    def __fspath__(self):
        """Comply with PathLike."""
        return os.fspath(self._file)

    def __repr__(self):
        """Representation of the object."""
        return '<FSFile "' + str(self._file) + '">'

    @property
    def fs(self):
        """Return the underlying private filesystem attribute."""
        return self._fs

    def open(self, *args, **kwargs):  # noqa: A003
        """Open the file.

        This is read-only.
        """
        fs_open_kwargs = self._update_with_fs_open_kwargs(kwargs)
        try:
            return self._fs.open(self._file, *args, **fs_open_kwargs)
        except AttributeError:
            return open(self._file, *args, **kwargs)

    def _update_with_fs_open_kwargs(self, user_kwargs):
        """Complement keyword arguments for opening a file via file system."""
        kwargs = user_kwargs.copy()
        kwargs.update(self._fs_open_kwargs)
        return kwargs

    def __lt__(self, other):
        """Implement ordering.

        Ordering is defined by the string representation of the filename,
        without considering the file system.
        """
        return os.fspath(self) < os.fspath(other)

    def __eq__(self, other):
        """Implement equality comparisons.

        Two FSFile instances are considered equal if they have the same
        filename and the same file system.
        """
        return (isinstance(other, FSFile) and
                self._file == other._file and
                self._fs == other._fs)

    def __hash__(self):
        """Implement hashing.

        Make FSFile objects hashable, so that they can be used in sets.  Some
        parts of satpy and perhaps others use sets of filenames (strings or
        pathlib.Path), or maybe use them as dictionary keys.  This requires
        them to be hashable.  To ensure FSFile can work as a drop-in
        replacement for strings of Path objects to represent the location of
        blob of data, FSFile should be hashable too.

        Returns the hash, computed from the hash of the filename and the hash
        of the filesystem.
        """
        try:
            fshash = hash(self._fs)
        except TypeError:  # fsspec < 0.8.8 for CachingFileSystem
            fshash = hash(pickle.dumps(self._fs))  # nosec B403
        return hash(self._file) ^ fshash


def _get_fs_open_kwargs(file):
    """Get keyword arguments for opening a file via file system.

    For example compression.
    """
    return {
        "compression": _get_compression(file)
    }


def _get_compression(file):
    try:
        return file.compression
    except AttributeError:
        return None


def open_file_or_filename(unknown_file_thing, mode=None):
    """Try to open the provided file "thing" if needed, otherwise return the filename or Path.

    This wraps the logic of getting something like an fsspec OpenFile object
    that is not directly supported by most reading libraries and making it
    usable. If a :class:`pathlib.Path` object or something that is not
    open-able is provided then that object is passed along. In the case of
    fsspec OpenFiles their ``.open()`` method is called and the result returned.

    """
    if isinstance(unknown_file_thing, pathlib.Path):
        f_obj = unknown_file_thing
    else:
        try:
            if mode is None:
                f_obj = unknown_file_thing.open()
            else:
                f_obj = unknown_file_thing.open(mode=mode)
        except AttributeError:
            f_obj = unknown_file_thing
    return f_obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for generic image (e.g. gif, png, jpg, tif, geotiff, ...).

Returns a dataset without calibration.  Includes coordinates if
available in the file (eg. geotiff).
If nodata values are present (and rasterio is able to read them), it
will be preserved as attribute ``_FillValue`` in the returned dataset.
In case that nodata values should be used to mask pixels (that have
equal values) with np.nan, it has to be enabled in the reader yaml
file (key ``nodata_handling`` per dataset with value ``"nan_mask"``).
"""

import logging

import dask.array as da
import numpy as np
import rasterio
import rioxarray  # noqa: F401  # need by xarray with the engine rasterio
import xarray as xr
from pyresample import utils

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()

BANDS = {1: ["L"],
         2: ["L", "A"],
         3: ["R", "G", "B"],
         4: ["R", "G", "B", "A"]}

NODATA_HANDLING_FILLVALUE = "fill_value"
NODATA_HANDLING_NANMASK = "nan_mask"

logger = logging.getLogger(__name__)


class GenericImageFileHandler(BaseFileHandler):
    """Handle reading of generic image files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize filehandler."""
        super(GenericImageFileHandler, self).__init__(
            filename, filename_info, filetype_info)
        self.finfo = filename_info
        try:
            self.finfo["end_time"] = self.finfo["start_time"]
        except KeyError:
            pass
        self.finfo["filename"] = self.filename
        self.file_content = {}
        self.area = None
        self.dataset_name = None
        self.read()

    def read(self):
        """Read the image."""
        dataset = rasterio.open(self.finfo["filename"])

        # Create area definition
        if hasattr(dataset, "crs") and dataset.crs is not None:
            self.area = utils.get_area_def_from_raster(dataset)

        # xarray use the engine 'rasterio' to open the file, but
        #   its actually rioxarray used in the backend.
        #   however, error is not explicit enough (see https://github.com/pydata/xarray/issues/7831)
        data = xr.open_dataset(self.finfo["filename"], engine="rasterio",
                               chunks={"band": 1, "y": CHUNK_SIZE, "x": CHUNK_SIZE}, mask_and_scale=False)["band_data"]
        if hasattr(dataset, "nodatavals"):
            # The nodata values for the raster bands
            # copied from https://github.com/pydata/xarray/blob/v2023.03.0/xarray/backends/rasterio_.py#L322-L326
            nodatavals = tuple(
                np.nan if nodataval is None else nodataval for nodataval in dataset.nodatavals
            )
            data.attrs["nodatavals"] = nodatavals

        attrs = data.attrs.copy()

        # Rename to Satpy convention
        data = data.rename({"band": "bands"})

        # Rename bands to [R, G, B, A], or a subset of those
        data["bands"] = BANDS[data.bands.size]

        data.attrs = attrs
        self.dataset_name = "image"
        self.file_content[self.dataset_name] = data

    def get_area_def(self, dsid):
        """Get area definition of the image."""
        if self.area is None:
            raise NotImplementedError("No CRS information available from image")
        return self.area

    @property
    def start_time(self):
        """Return start time."""
        return self.finfo["start_time"]

    @property
    def end_time(self):
        """Return end time."""
        return self.finfo["end_time"]

    def get_dataset(self, key, info):
        """Get a dataset from the file."""
        ds_name = self.dataset_name if self.dataset_name else key["name"]
        logger.debug("Reading '%s.'", ds_name)
        data = self.file_content[ds_name]

        # Mask data if necessary
        try:
            data = _mask_image_data(data, info)
        except ValueError as err:
            logger.warning(err)

        data.attrs.update(key.to_dict())
        data.attrs.update(info)
        return data


def _mask_image_data(data, info):
    """Mask image data if necessary.

    Masking is done if alpha channel is present or
    dataset 'nodata_handling' is set to 'nan_mask'.
    In the latter case even integer data is converted
    to float32 and masked with np.nan.
    """
    if data.bands.size in (2, 4):
        if not np.issubdtype(data.dtype, np.integer):
            raise ValueError("Only integer datatypes can be used as a mask.")
        mask = data.data[-1, :, :] == np.iinfo(data.dtype).min
        data = data.astype(np.float32)
        masked_data = da.stack([da.where(mask, np.nan, data.data[i, :, :])
                                for i in range(data.shape[0])])
        data.data = masked_data
        data = data.sel(bands=BANDS[data.bands.size - 1])
    elif hasattr(data, "nodatavals") and data.nodatavals:
        data = _handle_nodatavals(data, info.get("nodata_handling", NODATA_HANDLING_FILLVALUE))
    return data


def _handle_nodatavals(data, nodata_handling):
    """Mask data with np.nan or only set 'attr_FillValue'."""
    if nodata_handling == NODATA_HANDLING_NANMASK:
        # data converted to float and masked with np.nan
        data = data.astype(np.float32)
        masked_data = da.stack([da.where(data.data[i, :, :] == nodataval, np.nan, data.data[i, :, :])
                                for i, nodataval in enumerate(data.nodatavals)])
        data.data = masked_data
        data.attrs["_FillValue"] = np.nan
    elif nodata_handling == NODATA_HANDLING_FILLVALUE:
        # keep data as it is but set _FillValue attribute to provided
        # nodatavalue (first one as it has to be the same for all bands at least
        # in GeoTiff, see GDAL gtiff driver documentation)
        fill_value = data.nodatavals[0]
        if np.issubdtype(data.dtype, np.integer):
            fill_value = int(fill_value)
        data.attrs["_FillValue"] = fill_value
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the AWIPS Tiled writer."""

import datetime as dt
import logging
import os
import shutil
from glob import glob

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyproj import CRS

from satpy.tests.utils import assert_maximum_dask_computes

START_TIME = dt.datetime(2018, 1, 1, 12, 0, 0)
END_TIME = START_TIME + dt.timedelta(minutes=20)

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path
# - caplog


def _check_production_location(ds):
    if "production_site" in ds.attrs:
        prod_loc_name = "production_site"
    elif "production_location" in ds.attrs:
        prod_loc_name = "producton_location"
    else:
        return

    if prod_loc_name in ds.attrs:
        assert len(ds.attrs[prod_loc_name]) == 31


def check_required_properties(unmasked_ds, masked_ds):
    """Check various aspects of coordinates and attributes for correctness."""
    _check_scaled_x_coordinate_variable(unmasked_ds, masked_ds)
    _check_scaled_y_coordinate_variable(unmasked_ds, masked_ds)
    _check_required_common_attributes(unmasked_ds)


def _check_required_common_attributes(ds):
    """Check common properties of the created AWIPS tiles for validity."""
    for attr_name in ("tile_row_offset", "tile_column_offset",
                      "product_tile_height", "product_tile_width",
                      "number_product_tiles",
                      "product_rows", "product_columns"):
        assert attr_name in ds.attrs
    _check_production_location(ds)

    for data_arr in ds.data_vars.values():
        if data_arr.ndim == 0:
            # grid mapping variable
            assert "grid_mapping_name" in data_arr.attrs
            continue
        assert data_arr.encoding.get("zlib", False)
        assert "grid_mapping" in data_arr.attrs
        assert data_arr.attrs["grid_mapping"] in ds
        assert "units" in data_arr.attrs
        if data_arr.name != "DQF":
            assert data_arr.dtype == np.int16
            assert data_arr.attrs["_Unsigned"] == "true"


def _check_scaled_x_coordinate_variable(ds, masked_ds):
    assert "x" in ds.coords
    x_coord = ds.coords["x"]
    np.testing.assert_equal(np.diff(x_coord), 1)
    x_attrs = x_coord.attrs
    assert x_attrs.get("standard_name") == "projection_x_coordinate"
    assert x_attrs.get("units") == "meters"
    assert "scale_factor" in x_attrs
    assert x_attrs["scale_factor"] > 0
    assert "add_offset" in x_attrs

    unscaled_x = masked_ds.coords["x"].values
    assert (np.diff(unscaled_x) > 0).all()


def _check_scaled_y_coordinate_variable(ds, masked_ds):
    assert "y" in ds.coords
    y_coord = ds.coords["y"]
    np.testing.assert_equal(np.diff(y_coord), 1)
    y_attrs = y_coord.attrs
    assert y_attrs.get("standard_name") == "projection_y_coordinate"
    assert y_attrs.get("units") == "meters"
    assert "scale_factor" in y_attrs
    assert y_attrs["scale_factor"] < 0
    assert "add_offset" in y_attrs

    unscaled_y = masked_ds.coords["y"].values
    assert (np.diff(unscaled_y) < 0).all()


def _get_test_area(shape=(200, 100), crs=None, extents=None):
    from pyresample.geometry import AreaDefinition
    if crs is None:
        crs = CRS("+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs")
    if extents is None:
        extents = (-1000., -1500., 1000., 1500.)
    area_def = AreaDefinition(
        "test",
        "test",
        "test",
        crs,
        shape[1],
        shape[0],
        extents,
    )
    return area_def


def _get_test_data(shape=(200, 100), chunks=50):
    data = np.linspace(0., 1., shape[0] * shape[1], dtype=np.float32).reshape(shape)
    return da.from_array(data, chunks=chunks)


def _get_test_lcc_data(dask_arr, area_def, extra_attrs=None):
    from satpy.resample.base import _update_resampled_coords

    attrs = dict(
        name="test_ds",
        platform_name="PLAT",
        sensor="SENSOR",
        units="1",
        standard_name="toa_bidirectional_reflectance",
        area=area_def,
        start_time=START_TIME,
        end_time=END_TIME
    )
    if extra_attrs:
        attrs.update(extra_attrs)
    ds = xr.DataArray(
        dask_arr,
        dims=("y", "x") if dask_arr.ndim == 2 else ("bands", "y", "x"),
        attrs=attrs,
    )
    return _update_resampled_coords(ds, ds, area_def)


@pytest.fixture(autouse=True)
def short_prod_location(monkeypatch):
    """Force the environment variable used for "production_location" to a short string.

    If not set, the writer would use the current hostname, but some CI runners
    may have a long hostname and trigger warnings that we'd like to avoid.

    """
    monkeypatch.setenv("ORGANIZATION", "ABC")


class TestAWIPSTiledWriter:
    """Test basic functionality of AWIPS Tiled writer."""

    def test_init(self, tmp_path):
        """Test basic init method of writer."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        AWIPSTiledWriter(base_dir=str(tmp_path))

    @pytest.mark.parametrize("use_save_dataset",
                             [(False,), (True,)])
    @pytest.mark.parametrize(
        ("extra_attrs", "expected_filename"),
        [
            ({}, "TESTS_AII_PLAT_SENSOR_test_ds_TEST_T001_20180101_1200.nc"),
            ({"sensor": "viirs", "name": "I01"}, "TESTS_AII_PLAT_viirs_I01_TEST_T001_20180101_1200.nc"),
        ]
    )
    def test_basic_numbered_1_tile(self, extra_attrs, expected_filename, use_save_dataset, caplog, tmp_path):
        """Test creating a single numbered tile."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        data = _get_test_data()
        area_def = _get_test_area()
        input_data_arr = _get_test_lcc_data(data, area_def, extra_attrs)
        with assert_maximum_dask_computes(max_computes=1), \
                caplog.at_level(logging.DEBUG):
            w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
            if use_save_dataset:
                w.save_dataset(input_data_arr, sector_id="TEST", source_name="TESTS")
            else:
                w.save_datasets([input_data_arr], sector_id="TEST", source_name="TESTS")

        assert "no routine matching" not in caplog.text
        assert "Can't format string" not in caplog.text
        all_files = glob(os.path.join(str(tmp_path), "TESTS_AII*.nc"))
        assert len(all_files) == 1
        assert os.path.basename(all_files[0]) == expected_filename
        for fn in all_files:
            unmasked_ds = xr.open_dataset(fn, mask_and_scale=False)
            output_ds = xr.open_dataset(fn, mask_and_scale=True)
            check_required_properties(unmasked_ds, output_ds)
            scale_factor = output_ds["data"].encoding["scale_factor"]
            np.testing.assert_allclose(input_data_arr.values, output_ds["data"].data,
                                       atol=scale_factor * 0.75)

    def test_units_length_warning(self, tmp_path):
        """Test long 'units' warnings are raised."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        data = _get_test_data()
        area_def = _get_test_area()
        input_data_arr = _get_test_lcc_data(data, area_def)
        input_data_arr.attrs["units"] = "this is a really long units string"
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
        with pytest.warns(UserWarning, match=r".*this is a really long units string.*too long.*"):
            w.save_dataset(input_data_arr, sector_id="TEST", source_name="TESTS")

    @pytest.mark.parametrize(
        ("tile_count", "tile_size"),
        [
            ((3, 3), None),
            (None, (67, 34)),
            (None, None),
        ]
    )
    def test_basic_numbered_tiles(self, tile_count, tile_size, tmp_path):
        """Test creating a multiple numbered tiles."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        data = _get_test_data()
        area_def = _get_test_area()
        input_data_arr = _get_test_lcc_data(data, area_def)
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
        save_kwargs = dict(
            sector_id="TEST",
            source_name="TESTS",
            tile_count=tile_count,
            tile_size=tile_size,
            extra_global_attrs={"my_global": "TEST"}
        )
        should_error = tile_count is None and tile_size is None
        if should_error:
            with assert_maximum_dask_computes(max_computes=0), \
                 pytest.raises(ValueError, match=r"Either.*tile_count.*"):
                w.save_datasets([input_data_arr], **save_kwargs)
        else:
            with assert_maximum_dask_computes(max_computes=1):
                w.save_datasets([input_data_arr], **save_kwargs)

        all_files = glob(os.path.join(str(tmp_path), "TESTS_AII*.nc"))
        expected_num_files = 0 if should_error else 9
        assert len(all_files) == expected_num_files
        for fn in all_files:
            unmasked_ds = xr.open_dataset(fn, mask_and_scale=False)
            masked_ds = xr.open_dataset(fn, mask_and_scale=True)
            check_required_properties(unmasked_ds, masked_ds)
            assert unmasked_ds.attrs["my_global"] == "TEST"
            assert unmasked_ds.attrs["sector_id"] == "TEST"
            assert "physical_element" in unmasked_ds.attrs
            stime = input_data_arr.attrs["start_time"]
            assert unmasked_ds.attrs["start_date_time"] == stime.strftime("%Y-%m-%dT%H:%M:%S")

    def test_basic_lettered_tiles(self, tmp_path):
        """Test creating a lettered grid."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
        data = _get_test_data(shape=(2000, 1000), chunks=500)
        area_def = _get_test_area(shape=(2000, 1000),
                                  extents=(-1000000., -1500000., 1000000., 1500000.))
        ds = _get_test_lcc_data(data, area_def)
        # tile_count should be ignored since we specified lettered_grid
        with assert_maximum_dask_computes(max_computes=1):
            w.save_datasets([ds], sector_id="LCC", source_name="TESTS", tile_count=(3, 3), lettered_grid=True)
        all_files = glob(os.path.join(str(tmp_path), "TESTS_AII*.nc"))
        assert len(all_files) == 16
        for fn in all_files:
            unmasked_ds = xr.open_dataset(fn, mask_and_scale=False)
            masked_ds = xr.open_dataset(fn, mask_and_scale=True)
            check_required_properties(unmasked_ds, masked_ds)
            assert masked_ds.attrs["start_date_time"] == START_TIME.strftime("%Y-%m-%dT%H:%M:%S")

    def test_basic_lettered_tiles_diff_projection(self, tmp_path):
        """Test creating a lettered grid from data with differing projection.."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
        crs = CRS("+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. +lat_0=45 +lat_1=45 +units=m +no_defs")
        data = _get_test_data(shape=(2000, 1000), chunks=500)
        area_def = _get_test_area(shape=(2000, 1000), crs=crs,
                                  extents=(-1000000., -1500000., 1000000., 1500000.))
        ds = _get_test_lcc_data(data, area_def)
        # tile_count should be ignored since we specified lettered_grid
        with assert_maximum_dask_computes(max_computes=1):
            w.save_datasets([ds], sector_id="LCC", source_name="TESTS", tile_count=(3, 3), lettered_grid=True)
        all_files = sorted(glob(os.path.join(str(tmp_path), "TESTS_AII*.nc")))
        assert len(all_files) == 24
        assert "TC02" in all_files[0]  # the first tile should be TC02
        for fn in all_files:
            unmasked_ds = xr.open_dataset(fn, mask_and_scale=False)
            masked_ds = xr.open_dataset(fn, mask_and_scale=True)
            check_required_properties(unmasked_ds, masked_ds)
            assert masked_ds.attrs["start_date_time"] == START_TIME.strftime("%Y-%m-%dT%H:%M:%S")

    def test_lettered_tiles_update_existing(self, tmp_path):
        """Test updating lettered tiles with additional data."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        first_base_dir = os.path.join(str(tmp_path), "first")
        w = AWIPSTiledWriter(base_dir=first_base_dir, compress=True)
        shape = (2000, 1000)
        data = np.linspace(0., 1., shape[0] * shape[1], dtype=np.float32).reshape(shape)
        # pixels to be filled in later
        data[:, -200:] = np.nan
        data = da.from_array(data, chunks=500)
        area_def = _get_test_area(shape=(2000, 1000),
                                  extents=(-1000000., -1500000., 1000000., 1500000.))
        ds = _get_test_lcc_data(data, area_def)
        # tile_count should be ignored since we specified lettered_grid
        with assert_maximum_dask_computes(max_computes=1):
            w.save_datasets([ds], sector_id="LCC", source_name="TESTS", tile_count=(3, 3), lettered_grid=True)
        all_files = sorted(glob(os.path.join(first_base_dir, "TESTS_AII*.nc")))
        assert len(all_files) == 16
        first_files = []
        second_base_dir = os.path.join(str(tmp_path), "second")
        os.makedirs(second_base_dir)
        for fn in all_files:
            new_fn = fn.replace(first_base_dir, second_base_dir)
            shutil.copy(fn, new_fn)
            first_files.append(new_fn)

        # Second writing/updating
        # Area is about 100 pixels to the right
        area_def2 = _get_test_area(shape=(2000, 1000),
                                   extents=(-800000., -1500000., 1200000., 1500000.))
        data2 = np.linspace(0., 1., 2000000, dtype=np.float32).reshape((2000, 1000))
        # a gap at the beginning where old values remain
        data2[:, :200] = np.nan
        # a gap at the end where old values remain
        data2[:, -400:-300] = np.nan
        data2 = da.from_array(data2, chunks=500)
        ds2 = _get_test_lcc_data(data2, area_def2)
        w = AWIPSTiledWriter(base_dir=second_base_dir, compress=True)
        with assert_maximum_dask_computes(max_computes=1):
            w.save_datasets([ds2], sector_id="LCC", source_name="TESTS", tile_count=(3, 3), lettered_grid=True)
        all_files = glob(os.path.join(second_base_dir, "TESTS_AII*.nc"))
        # 16 original tiles + 4 new tiles
        assert len(all_files) == 20

        # these tiles should be the right-most edge of the first image
        first_right_edge_files = [x for x in first_files if "P02" in x or "P04" in x or "V02" in x or "V04" in x]
        for new_file in first_right_edge_files:
            orig_file = new_file.replace(second_base_dir, first_base_dir)
            orig_nc = xr.open_dataset(orig_file)
            orig_data = orig_nc["data"].values
            if not np.isnan(orig_data).any():
                # we only care about the tiles that had NaNs originally
                continue

            new_nc = xr.open_dataset(new_file)
            new_data = new_nc["data"].values
            # there should be at least some areas of the file
            # that old data was present and hasn't been replaced
            np.testing.assert_allclose(orig_data[:, :20], new_data[:, :20])
            # it isn't exactly 200 because the tiles aren't aligned with the
            # data (the left-most tile doesn't have data until some columns
            # in), but it should be at least that many columns
            assert np.isnan(orig_data[:, 200:]).all()
            assert not np.isnan(new_data[:, 200:]).all()

    def test_lettered_tiles_sector_ref(self, tmp_path):
        """Test creating a lettered grid using the sector as reference."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
        data = _get_test_data(shape=(2000, 1000), chunks=500)
        area_def = _get_test_area(shape=(2000, 1000),
                                  extents=(-1000000., -1500000., 1000000., 1500000.))
        ds = _get_test_lcc_data(data, area_def)
        with assert_maximum_dask_computes(max_computes=1):
            w.save_datasets([ds], sector_id="LCC", source_name="TESTS",
                            lettered_grid=True, use_sector_reference=True,
                            use_end_time=True)
        all_files = glob(os.path.join(str(tmp_path), "TESTS_AII*.nc"))
        assert len(all_files) == 16
        for fn in all_files:
            unmasked_ds = xr.open_dataset(fn, mask_and_scale=False)
            masked_ds = xr.open_dataset(fn, mask_and_scale=True)
            check_required_properties(unmasked_ds, masked_ds)
            expected_start = (START_TIME + dt.timedelta(minutes=20)).strftime("%Y-%m-%dT%H:%M:%S")
            assert masked_ds.attrs["start_date_time"] == expected_start

    def test_lettered_tiles_no_fit(self, tmp_path):
        """Test creating a lettered grid with no data overlapping the grid."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
        data = _get_test_data(shape=(2000, 1000), chunks=500)
        area_def = _get_test_area(shape=(2000, 1000),
                                  extents=(4000000., 5000000., 5000000., 6000000.))
        ds = _get_test_lcc_data(data, area_def)
        with assert_maximum_dask_computes(max_computes=1):
            w.save_datasets([ds], sector_id="LCC", source_name="TESTS", tile_count=(3, 3), lettered_grid=True)
        # No files created
        all_files = glob(os.path.join(str(tmp_path), "TESTS_AII*.nc"))
        assert not all_files

    def test_lettered_tiles_no_valid_data(self, tmp_path):
        """Test creating a lettered grid with no valid data."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
        data = da.full((2000, 1000), np.nan, chunks=500, dtype=np.float32)
        area_def = _get_test_area(shape=(2000, 1000),
                                  extents=(-1000000., -1500000., 1000000., 1500000.))
        ds = _get_test_lcc_data(data, area_def)
        with assert_maximum_dask_computes(max_computes=1):
            w.save_datasets([ds], sector_id="LCC", source_name="TESTS", tile_count=(3, 3), lettered_grid=True)
        # No files created - all NaNs should result in no tiles being created
        all_files = glob(os.path.join(str(tmp_path), "TESTS_AII*.nc"))
        assert not all_files

    def test_lettered_tiles_bad_filename(self, tmp_path):
        """Test creating a lettered grid with a bad filename."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True, filename="{Bad Key}.nc")
        data = _get_test_data(shape=(2000, 1000), chunks=500)
        area_def = _get_test_area(shape=(2000, 1000),
                                  extents=(-1000000., -1500000., 1000000., 1500000.))
        ds = _get_test_lcc_data(data, area_def)
        with assert_maximum_dask_computes(max_computes=1), \
                pytest.raises(KeyError):
            w.save_datasets([ds],
                            sector_id="LCC",
                            source_name="TESTS",
                            tile_count=(3, 3),
                            lettered_grid=True)

    def test_basic_numbered_tiles_rgb(self, tmp_path):
        """Test creating a multiple numbered tiles with RGB."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter
        w = AWIPSTiledWriter(base_dir=str(tmp_path), compress=True)
        data = da.from_array(np.linspace(0., 1., 60000, dtype=np.float32).reshape((3, 200, 100)), chunks=50)
        area_def = _get_test_area()
        ds = _get_test_lcc_data(data, area_def)
        ds = ds.rename(dict((old, new) for old, new in zip(ds.dims, ["bands", "y", "x"])))
        ds.coords["bands"] = ["R", "G", "B"]

        with assert_maximum_dask_computes(max_computes=1):
            w.save_datasets([ds], sector_id="TEST", source_name="TESTS", tile_count=(3, 3))
        chan_files = glob(os.path.join(str(tmp_path), "TESTS_AII*test_ds_R*.nc"))
        all_files = chan_files[:]
        assert len(chan_files) == 9
        chan_files = glob(os.path.join(str(tmp_path), "TESTS_AII*test_ds_G*.nc"))
        all_files.extend(chan_files)
        assert len(chan_files) == 9
        chan_files = glob(os.path.join(str(tmp_path), "TESTS_AII*test_ds_B*.nc"))
        assert len(chan_files) == 9
        all_files.extend(chan_files)
        for fn in all_files:
            unmasked_ds = xr.open_dataset(fn, mask_and_scale=False)
            masked_ds = xr.open_dataset(fn, mask_and_scale=True)
            check_required_properties(unmasked_ds, masked_ds)

    @pytest.mark.parametrize(
        "sector",
        ["C",
         "F"]
    )
    @pytest.mark.parametrize(
        "extra_kwargs",
        [
            {},
            {"environment_prefix": "AA"},
            {"environment_prefix": "BB", "filename": "{environment_prefix}_{name}_GLM_T{tile_number:04d}.nc"},
        ]
    )
    def test_multivar_numbered_tiles_glm(self, sector, extra_kwargs, tmp_path, monkeypatch):
        """Test creating a tiles with multiple variables."""
        from satpy.writers.awips_tiled import AWIPSTiledWriter

        monkeypatch.setenv("ORGANIZATION", "1" * 50)
        w = AWIPSTiledWriter(base_dir=tmp_path, compress=True)
        data = _get_test_data()
        area_def = _get_test_area()
        ds1 = _get_test_lcc_data(data, area_def)
        ds1.attrs.update(
            dict(
                name="total_energy",
                platform_name="GOES-17",
                sensor="SENSOR",
                units="1",
                scan_mode="M3",
                scene_abbr=sector,
                platform_shortname="G17"
            )
        )
        ds2 = ds1.copy()
        ds2.attrs.update({
            "name": "flash_extent_density",
        })
        ds3 = ds1.copy()
        ds3.attrs.update({
            "name": "average_flash_area",
        })
        dqf = ds1.copy()
        dqf = (dqf * 255).astype(np.uint8)
        dqf.attrs = ds1.attrs.copy()
        dqf.attrs.update({
            "name": "DQF",
            "_FillValue": 1,
        })

        with assert_maximum_dask_computes(max_computes=1), \
                pytest.warns(UserWarning, match="Production location attribute "):
            w.save_datasets([ds1, ds2, ds3, dqf], sector_id="TEST", source_name="TESTS",
                            tile_count=(3, 3), template="glm_l2_rad{}".format(sector.lower()),
                            **extra_kwargs)
        fn_glob = self._get_glm_glob_filename(extra_kwargs)
        all_files = glob(os.path.join(str(tmp_path), fn_glob))
        assert len(all_files) == 9
        for fn in all_files:
            unmasked_ds = xr.open_dataset(fn, mask_and_scale=False)
            masked_ds = xr.open_dataset(fn, mask_and_scale=True)
            check_required_properties(unmasked_ds, masked_ds)
            if sector == "C":
                assert masked_ds.attrs["time_coverage_end"] == END_TIME.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            else:  # 'F'
                assert masked_ds.attrs["time_coverage_end"] == END_TIME.strftime("%Y-%m-%dT%H:%M:%SZ")

    @staticmethod
    def _get_glm_glob_filename(extra_kwargs):
        if "filename" in extra_kwargs:
            return "BB*_GLM*.nc"
        elif "environment_prefix" in extra_kwargs:
            return "AA*_GLM*.nc"
        return "DR*_GLM*.nc"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.viirs_l1b module."""

import datetime as dt
import os
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)

class FakeNetCDF4FileHandlerDay(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    M_REFL_BANDS = [f"M{band_num:02d}" for band_num in range(1, 12)]
    M_BT_BANDS = [f"M{band_num:02d}" for band_num in range(12, 17)]
    M_BANDS = M_REFL_BANDS + M_BT_BANDS
    M_BANDS_FLAG = [f"{band}_quality_flags" for band in M_BANDS]
    I_REFL_BANDS = [f"I{band_num:02d}" for band_num in range(1, 4)]
    I_BT_BANDS = [f"I{band_num:02d}" for band_num in range(4, 6)]
    I_BANDS = I_REFL_BANDS + I_BT_BANDS
    I_BANDS_FLAG = [f"{band}_quality_flags" for band in I_BANDS]

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        date = filename_info.get("start_time", dt.datetime(2016, 1, 1, 12, 0, 0))
        file_type = filename[:5].lower()
        num_lines = DEFAULT_FILE_SHAPE[0]
        num_pixels = DEFAULT_FILE_SHAPE[1]
        num_scans = 5
        num_luts = DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1]
        file_content = {
            "/dimension/number_of_scans": num_scans,
            "/dimension/number_of_lines": num_lines,
            "/dimension/number_of_pixels": num_pixels,
            "/dimension/number_of_LUT_values": num_luts,
            "/attr/time_coverage_start": date.strftime("%Y-%m-%dT%H:%M:%S.000Z"),
            "/attr/time_coverage_end": (date + dt.timedelta(minutes=6)).strftime("%Y-%m-%dT%H:%M:%S.000Z"),
            "/attr/orbit_number": 26384,
            "/attr/instrument": "VIIRS",
            "/attr/platform": "Suomi-NPP",
            "/attr/DayNightFlag": "Day",
            "/attr/startDirection": "Descending",
            "/attr/endDirection": "Ascending",
        }
        self._fill_contents_with_default_data(file_content, file_type)
        self._set_dataset_specific_metadata(file_content)
        convert_file_content_to_data_array(file_content)
        return file_content

    def _fill_contents_with_default_data(self, file_content, file_type):
        """Fill file contents with default data."""
        if file_type.startswith("vgeo"):
            file_content["/attr/OrbitNumber"] = file_content.pop("/attr/orbit_number")
            file_content["geolocation_data/latitude"] = DEFAULT_LAT_DATA
            file_content["geolocation_data/longitude"] = DEFAULT_LON_DATA
            file_content["geolocation_data/solar_zenith"] = DEFAULT_LON_DATA
            file_content["geolocation_data/solar_azimuth"] = DEFAULT_LON_DATA
            file_content["geolocation_data/sensor_zenith"] = DEFAULT_LON_DATA
            file_content["geolocation_data/sensor_azimuth"] = DEFAULT_LON_DATA
            if file_type.endswith("d"):
                file_content["geolocation_data/lunar_zenith"] = DEFAULT_LON_DATA
                file_content["geolocation_data/lunar_azimuth"] = DEFAULT_LON_DATA
        elif file_type == "vl1bm":
            for m_band in self.M_BANDS:
                file_content[f"observation_data/{m_band}"] = DEFAULT_FILE_DATA
            for m_band_flag in self.M_BANDS_FLAG:
                file_content[f"observation_data/{m_band_flag}"] = DEFAULT_FILE_DATA
        elif file_type == "vl1bi":
            for i_band in self.I_BANDS:
                file_content[f"observation_data/{i_band}"] = DEFAULT_FILE_DATA
            for i_band_flag in self.I_BANDS_FLAG:
                file_content[f"observation_data/{i_band_flag}"] = DEFAULT_FILE_DATA
        elif file_type == "vl1bd":
            file_content["observation_data/DNB_observations"] = DEFAULT_FILE_DATA
            file_content["observation_data/DNB_observations/attr/units"] = "Watts/cm^2/steradian"

    @staticmethod
    def _set_dataset_specific_metadata(file_content):
        """Set dataset-specific metadata."""
        for k in list(file_content.keys()):
            if not k.startswith("observation_data") and not k.startswith("geolocation_data"):
                continue
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            if k[-3:] in ["M12", "M13", "M14", "M15", "M16", "I04", "I05"]:
                file_content[k + "_brightness_temperature_lut"] = DEFAULT_FILE_DATA.ravel()
                file_content[k + "_brightness_temperature_lut/attr/units"] = "Kelvin"
                file_content[k + "_brightness_temperature_lut/attr/valid_min"] = 0
                file_content[k + "_brightness_temperature_lut/attr/valid_max"] = 65534
                file_content[k + "_brightness_temperature_lut/attr/_FillValue"] = 65535
                file_content[k + "/attr/units"] = "Watts/meter^2/steradian/micrometer"
            elif k[-3:] in ["M01", "M02", "M03", "M04", "M05", "M06", "M07", "M08",
                            "M09", "M10", "M11", "I01", "I02", "I03"]:
                file_content[k + "/attr/radiance_units"] = "Watts/meter^2/steradian/micrometer"
                file_content[k + "/attr/radiance_scale_factor"] = 1.1
                file_content[k + "/attr/radiance_add_offset"] = 0.1
            elif k.endswith("longitude"):
                file_content[k + "/attr/units"] = "degrees_east"
            elif k.endswith("latitude"):
                file_content[k + "/attr/units"] = "degrees_north"
            elif k.endswith("zenith") or k.endswith("azimuth"):
                file_content[k + "/attr/units"] = "degrees"

            if k.endswith("quality_flags"):
                file_content[k + "/attr/units"] = "none"
            else:
                file_content[k + "/attr/valid_min"] = 0
                file_content[k + "/attr/valid_max"] = 65534
                file_content[k + "/attr/_FillValue"] = 65535
                file_content[k + "/attr/scale_factor"] = 1.1
                file_content[k + "/attr/add_offset"] = 0.1


class FakeNetCDF4FileHandlerNight(FakeNetCDF4FileHandlerDay):
    """Same as the day file handler, but some day-only bands are missing.

    This matches what happens in real world files where reflectance bands
    are removed in night data to save space.

    """

    M_BANDS = FakeNetCDF4FileHandlerDay.M_BT_BANDS
    I_BANDS = FakeNetCDF4FileHandlerDay.I_BT_BANDS


class TestVIIRSL1BReaderDay:
    """Test VIIRS L1B Reader."""

    yaml_file = "viirs_l1b.yaml"
    fake_cls = FakeNetCDF4FileHandlerDay
    has_reflectance_bands = True

    def setup_method(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_l1b import VIIRSL1BFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(VIIRSL1BFileHandler, "__bases__", (self.fake_cls,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def teardown_method(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BM_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_available_datasets_m_bands(self):
        """Test available datasets for M band files."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BM_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOM_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        r.create_filehandlers(loadables)
        avail_names = r.available_dataset_names
        angles = {"satellite_azimuth_angle", "satellite_zenith_angle", "solar_azimuth_angle", "solar_zenith_angle"}
        geo = {"m_lon", "m_lat"}
        assert set(avail_names) == set(self.fake_cls.M_BANDS) | angles | geo | set(self.fake_cls.M_BANDS_FLAG)

    def test_load_every_m_band_bt(self):
        """Test loading all M band brightness temperatures."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BM_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOM_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["M12",
                           "M13",
                           "M14",
                           "M15",
                           "M16"])
        assert len(datasets) == 5
        for v in datasets.values():
            assert v.attrs["calibration"] == "brightness_temperature"
            assert v.attrs["units"] == "K"
            assert v.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lons.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lats.attrs["rows_per_scan"] == 2
            assert v.attrs["sensor"] == "viirs"
            assert "scale_factor" not in v.attrs
            assert "add_offset" not in v.attrs

    def test_load_every_m_band_refl(self):
        """Test loading all M band reflectances."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BM_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOM_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["M01",
                           "M02",
                           "M03",
                           "M04",
                           "M05",
                           "M06",
                           "M07",
                           "M08",
                           "M09",
                           "M10",
                           "M11"])
        assert len(datasets) == (11 if self.has_reflectance_bands else 0)
        for v in datasets.values():
            assert v.attrs["calibration"] == "reflectance"
            assert v.attrs["units"] == "%"
            assert v.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lons.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lats.attrs["rows_per_scan"] == 2
            assert v.attrs["sensor"] == "viirs"
            assert "scale_factor" not in v.attrs
            assert "add_offset" not in v.attrs

    def test_load_every_m_band_rad(self):
        """Test loading all M bands as radiances."""
        from satpy.readers.core.loading import load_reader
        from satpy.tests.utils import make_dataid
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BM_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOM_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load([make_dataid(name="M01", calibration="radiance"),
                           make_dataid(name="M02", calibration="radiance"),
                           make_dataid(name="M03", calibration="radiance"),
                           make_dataid(name="M04", calibration="radiance"),
                           make_dataid(name="M05", calibration="radiance"),
                           make_dataid(name="M06", calibration="radiance"),
                           make_dataid(name="M07", calibration="radiance"),
                           make_dataid(name="M08", calibration="radiance"),
                           make_dataid(name="M09", calibration="radiance"),
                           make_dataid(name="M10", calibration="radiance"),
                           make_dataid(name="M11", calibration="radiance"),
                           make_dataid(name="M12", calibration="radiance"),
                           make_dataid(name="M13", calibration="radiance"),
                           make_dataid(name="M14", calibration="radiance"),
                           make_dataid(name="M15", calibration="radiance"),
                           make_dataid(name="M16", calibration="radiance")])
        assert len(datasets) == (16 if self.has_reflectance_bands else 5)
        for v in datasets.values():
            assert v.attrs["calibration"] == "radiance"
            assert v.attrs["units"] == "W m-2 um-1 sr-1"
            assert v.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lons.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lats.attrs["rows_per_scan"] == 2
            assert v.attrs["sensor"] == "viirs"
            assert v.attrs["day_night"] == "Day"
            assert v.attrs["orbital_parameters"]["start_direction"] == "Descending"
            assert v.attrs["orbital_parameters"]["end_direction"] == "Ascending"
            assert v.attrs["orbital_parameters"]["start_orbit"] == 26384
            assert v.attrs["orbital_parameters"]["end_orbit"] == 26384
            assert "scale_factor" not in v.attrs
            assert "add_offset" not in v.attrs

    def test_load_i_band_angles(self):
        """Test loading all M bands as radiances."""
        from satpy.readers.core.loading import load_reader
        from satpy.tests.utils import make_dataid
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BI_snpp_d20161130_t012400_c20161130054822.nc",
            "VL1BM_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOI_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOM_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load([
            make_dataid(name="satellite_zenith_angle"),
            make_dataid(name="satellite_azimuth_angle"),
            make_dataid(name="solar_azimuth_angle"),
            make_dataid(name="solar_zenith_angle"),
        ])
        assert len(datasets) == 4
        for v in datasets.values():
            assert v.attrs["resolution"] == 371
            assert v.attrs["sensor"] == "viirs"
            assert "scale_factor" not in v.attrs
            assert "add_offset" not in v.attrs

    def test_load_dnb_radiance(self):
        """Test loading the main DNB dataset."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BD_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOD_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["DNB"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["calibration"] == "radiance"
            assert v.attrs["units"] == "W m-2 sr-1"
            assert v.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lons.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lats.attrs["rows_per_scan"] == 2
            assert v.attrs["sensor"] == "viirs"
            assert "scale_factor" not in v.attrs
            assert "add_offset" not in v.attrs

    def test_load_dnb_angles(self):
        """Test loading all DNB angle datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BD_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOD_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["dnb_solar_zenith_angle",
                           "dnb_solar_azimuth_angle",
                           "dnb_satellite_zenith_angle",
                           "dnb_satellite_azimuth_angle",
                           "dnb_lunar_zenith_angle",
                           "dnb_lunar_azimuth_angle",
                           ])
        assert len(datasets) == 6
        for v in datasets.values():
            assert v.attrs["units"] == "degrees"
            assert v.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lons.attrs["rows_per_scan"] == 2
            assert v.attrs["area"].lats.attrs["rows_per_scan"] == 2
            assert v.attrs["sensor"] == "viirs"
            assert "scale_factor" not in v.attrs
            assert "add_offset" not in v.attrs

    def test_available_datasets_quality_flags(self):
        """Test that <name>_quality_flags datasets are available when present."""
        from satpy.readers.core.loading import load_reader

        # Load the reader with fake file contents
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "VL1BI_snpp_d20161130_t012400_c20161130054822.nc",
            "VL1BM_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOI_snpp_d20161130_t012400_c20161130054822.nc",
            "VGEOM_snpp_d20161130_t012400_c20161130054822.nc",
        ])
        r.create_filehandlers(loadables)

        datasets = r.load(["I01_quality_flags",
                           "M01_quality_flags",
                           ])
        assert len(datasets) == 2
        for v in datasets.values():
            assert v.attrs["units"] == "1"
            assert v.attrs["sensor"] == "viirs"
            assert "scale_factor" not in v.attrs
            assert "add_offset" not in v.attrs


class TestVIIRSL1BReaderDayNight(TestVIIRSL1BReaderDay):
    """Test VIIRS L1b with night data.

    Night data files don't have reflectance bands in them.

    """

    fake_cls = FakeNetCDF4FileHandlerNight
    has_reflectance_bands = False
